#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
from datetime import datetime
from time import sleep
from typing import List

from bitcoin_nostr_chat.connected_devices.connected_devices import (
    TrustedDevice,
    short_key,
)
from bitcoin_nostr_chat.nostr import BitcoinDM, ChatLabel
from nostr_sdk import PublicKey
from PyQt6.QtCore import QObject

from bitcoin_safe.gui.qt.sync_tab import SyncTab

logger = logging.getLogger(__name__)
from bitcoin_nostr_chat.nostr_sync import Data, DataType

from bitcoin_safe.labels import Labels, LabelType
from bitcoin_safe.signals import UpdateFilter, UpdateFilterReason, WalletSignals


class LabelSyncer(QObject):
    def __init__(self, labels: Labels, sync_tab: SyncTab, wallet_signals: WalletSignals) -> None:
        super().__init__()
        self.labels = labels
        self.sync_tab = sync_tab
        self.nostr_sync = sync_tab.nostr_sync
        self.wallet_signals = wallet_signals

        self.apply_own_labels = True

        self.nostr_sync.signal_label_bip329_received.connect(self.on_nostr_label_bip329_received)
        self.nostr_sync.signal_add_trusted_device.connect(self.on_add_trusted_device)
        self.wallet_signals.updated.connect(self.on_labels_updated)

    @staticmethod
    def chunk_lines(lines: List[str], max_len: int = 60_000) -> List[List[str]]:
        len_of_lines = [len(line) for line in lines]  # Calculate the length of each line

        # Determine split points
        split_indices = []
        current_length = 0

        for i, line_length in enumerate(len_of_lines):
            if current_length + line_length > max_len:
                split_indices.append(i)
                current_length = line_length  # Start new chunk with current line
            else:
                current_length += line_length + 1  # +1 for the newline, included in the next chunk

        # Use split indices to construct chunks
        chunks = []
        start_index = 0

        for index in split_indices:
            chunks.append(lines[start_index:index])
            start_index = index

        # Add the final chunk
        if start_index < len(lines):
            chunks.append(lines[start_index:])

        return chunks

    def get_chunked_bitcoin_data(self, refs: List[str]) -> List[Data]:
        lines = self.labels.dumps_data_jsonline_list(refs=refs)
        chunks = self.chunk_lines(lines, max_len=60_000)
        return [Data(data="\n".join(chunk), data_type=DataType.LabelsBip329) for chunk in chunks]

    def on_add_trusted_device(self, trusted_device: TrustedDevice) -> None:
        if not self.sync_tab.enabled():
            return
        logger.debug(f"on_add_trusted_device")

        # send entire label data
        refs = list(self.labels.data.keys())

        for bitcoin_data in self.get_chunked_bitcoin_data(refs):
            self.nostr_sync.group_chat.dm_connection.send(
                BitcoinDM(
                    event=None,
                    label=ChatLabel.SingleRecipient,
                    description="",
                    data=bitcoin_data,
                    created_at=datetime.now(),
                ),
                PublicKey.from_bech32(trusted_device.pub_key_bech32),
            )
        logger.info(f"Sent all labels to trusted device {short_key( trusted_device.pub_key_bech32)}")

    def on_nostr_label_bip329_received(self, data: Data, author: PublicKey) -> None:
        if not self.sync_tab.enabled():
            return

        if data.data_type != DataType.LabelsBip329:
            logger.debug(f"on_nostr_label_bip329_received received wrong type {type(data)}")
            return

        if self.sync_tab.nostr_sync.is_me(author) and not self.apply_own_labels:
            logger.debug(f"on_nostr_label_bip329_received do not apply laybels from myself {author}")
            return

        changed_labels = self.labels.import_dumps_data(data.data)
        if not changed_labels:
            logger.debug(f"no labels changed in on_nostr_label_bip329_received")
            return
        logger.info(f"on_nostr_label_bip329_received applied {len(changed_labels)} labels: {changed_labels} ")

        addresses: List[str] = []
        txids: List[str] = []
        for label in changed_labels.values():
            if label.type == LabelType.addr:
                addresses.append(label.ref)
            elif label.type == LabelType.tx:
                txids.append(label.ref)

        new_categories = [label.category for label in changed_labels.values()]
        update_filter = UpdateFilter(
            addresses=addresses,
            txids=txids,
            categories=new_categories,
            reason=UpdateFilterReason.SourceLabelSyncer,
        )
        #  make the wallet add new addresses
        self.wallet_signals.updated.emit(update_filter)
        logger.info(
            f"{self.__class__.__name__}: Received {len(addresses)} addresses, {len(txids)} txids, {len(new_categories)} categories  from {short_key(author.to_bech32())}"
        )

    def on_labels_updated(self, update_filter: UpdateFilter) -> None:
        if not self.sync_tab.enabled():
            return
        if update_filter.reason == UpdateFilterReason.SourceLabelSyncer:
            logger.debug("on_labels_updated: Do nothing because update_filter was sent from here.")
            return
        if update_filter.refresh_all:
            logger.debug("on_labels_updated: Do nothing on refresh_all.")
            return

        should_update = False
        if should_update or update_filter.refresh_all:
            should_update = True
        if should_update or update_filter.addresses:
            should_update = True
        if should_update or update_filter.categories:
            should_update = True

        if not should_update:
            return

        logger.debug(f"{self.__class__.__name__} update_with_filter {update_filter}")

        refs = list(update_filter.addresses) + list(update_filter.txids)
        if not refs:
            return

        for bitcoin_data in self.get_chunked_bitcoin_data(refs):
            self.nostr_sync.group_chat.send(
                BitcoinDM(
                    event=None,
                    label=ChatLabel.GroupChat,
                    description="",
                    data=bitcoin_data,
                    created_at=datetime.now(),
                )
            )
        logger.info(
            f"{self.__class__.__name__}: Sent {len(update_filter.addresses)} addresses, {len(update_filter.txids)} txids to {[short_key(m.to_bech32()) for m in  self.nostr_sync.group_chat.members]}"
        )

    def send_all_labels_to_myself(self):
        if not self.sync_tab.enabled():
            return
        logger.debug(f"send_all_labels_to_myself")

        # send entire label data
        refs = list(self.labels.data.keys())

        my_key = self.nostr_sync.group_chat.dm_connection.async_dm_connection.keys.public_key()
        for bitcoin_data in self.get_chunked_bitcoin_data(refs):
            self.nostr_sync.group_chat.dm_connection.send(
                BitcoinDM(
                    event=None,
                    label=ChatLabel.SingleRecipient,
                    description="",
                    data=bitcoin_data,
                    created_at=datetime.now(),
                ),
                my_key,
            )
        # sleep here to give the relays time to receive the message
        # but if the relays fail, then better fail sending the messages,
        # than blocking the wallet from closing
        sleep(0.2)



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from typing import Optional

from PyQt6.QtWidgets import (
    QApplication,
    QLabel,
    QMainWindow,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from bitcoin_safe.gui.qt.unique_deque import UniqueDeque


class HistTabWidget(QTabWidget):
    """Stores the closing activation history of the tabs and upon close, activates the last active one.

    Args:
        QTabWidget: Inherits from QTabWidget.
    """

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self._tab_history: UniqueDeque[int] = UniqueDeque(maxlen=100000)  # History of activated tab indices
        self.currentChanged.connect(self.on_current_changed)

    def on_current_changed(self, index: int) -> None:
        """Updates the tab history when the current tab is changed.

        Args:
            index (int): The index of the newly activated tab.
        """
        if index >= 0:
            self._tab_history.append(index)

    def remove_tab_from_history(self, index: int) -> None:
        """Handles the tab close request, updating history and setting the last active tab.

        Args:
            index (int): The index of the tab that is being closed.
        """
        # Remove the closed tab from history and adjust the indices
        if index in self._tab_history:
            self._tab_history = UniqueDeque([i for i in self._tab_history if i != index])
        self._tab_history = UniqueDeque([i - 1 if i > index else i for i in self._tab_history])

    def get_last_active_tab(self) -> int:
        if len(self._tab_history) >= 2:
            return self._tab_history[-2]
        elif len(self._tab_history) >= 1:
            return self._tab_history[-1]
        return self.currentIndex()

    def jump_to_last_active_tab(self) -> None:
        """Sets the current tab to the last active one from history or to the first tab if history is empty."""
        index = self.get_last_active_tab()
        if index >= 0:
            self.setCurrentIndex(index)

    def removeTab(self, index: int) -> None:
        self.remove_tab_from_history(index)
        return super().removeTab(index)


if __name__ == "__main__":

    class MainWindow(QMainWindow):
        def __init__(self):
            super().__init__()
            self.tab_widget = HistTabWidget()
            self.tab_widget.setTabsClosable(True)

            def remove(index):
                self.tab_widget.jump_to_last_active_tab()
                self.tab_widget.removeTab(index)

            self.tab_widget.tabCloseRequested.connect(remove)
            self.tab_widget.currentChanged.connect(
                lambda: print(f"on_currentChanged = {self.tab_widget._tab_history}")
            )
            self.setCentralWidget(self.tab_widget)
            # Adding example tabs
            for i in range(5):
                tab = QWidget()
                layout = QVBoxLayout()
                label = QLabel(f"Content of tab {i + 1}")
                layout.addWidget(label)
                tab.setLayout(layout)
                self.tab_widget.addTab(tab, f"Tab {i + 1}")

            self.setGeometry(300, 300, 400, 300)

    app = QApplication([])
    window = MainWindow()
    window.show()
    app.exec()



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
from typing import Dict, Generic, Type, TypeVar

from bitcoin_safe.gui.qt.histtabwidget import HistTabWidget

logger = logging.getLogger(__name__)

from typing import Dict, Generic, Type, TypeVar

from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import QApplication, QWidget

T = TypeVar("T")


class DataTabWidget(Generic[T], HistTabWidget):
    def __init__(self, data_class: Type[T], parent=None) -> None:
        super().__init__(parent)
        self._data_class = data_class
        self._tab_data: Dict[QWidget, T] = {}

    def setTabData(self, widget: QWidget, data: T) -> None:
        self._tab_data[widget] = data

    def tabData(self, index: int) -> T | None:
        tab = self.widget(index)
        if not tab:
            return None
        return self._tab_data[tab]

    def get_data_for_tab(self, tab: QWidget) -> T:
        return self._tab_data[tab]

    def getCurrentTabData(self) -> T | None:
        current_widget = self.currentWidget()
        if not current_widget:
            return None
        return self._tab_data[current_widget]

    def getAllTabData(self) -> Dict[QWidget, T]:
        widgets_raw = [self.widget(i) for i in range(self.count())]
        widgets = [w for w in widgets_raw if w]
        return {widget: self.get_data_for_tab(widget) for widget in widgets}

    def clearTabData(self) -> None:
        self._tab_data.clear()

    def clear(self) -> None:
        """Override the clear method to also clear the tab data."""
        super().clear()
        self._tab_data.clear()

    def addTab(  # type: ignore[override]
        self, widget: QWidget, icon: QIcon | None = None, description: str = "", data: T | None = None
    ) -> int:  # type: ignore[override]
        if icon:
            index = super().addTab(widget, icon, description)
        else:
            index = super().addTab(widget, description)
        if data is not None:
            self.setTabData(widget, data)
        return index

    def insertTab(  # type: ignore[override]
        self, index: int, widget: QWidget, data: T, icon: QIcon | None = None, description: str = ""
    ) -> int:  # type: ignore[override]
        if icon:
            new_index = super().insertTab(index, widget, icon, description)
        else:
            new_index = super().insertTab(index, widget, description)
        if data is not None:
            self.setTabData(widget, data)
        return new_index

    def add_tab(
        self,
        tab: QWidget,
        icon: QIcon | None,
        description: str,
        data: T,
        position: int | None = None,
        focus: bool = False,
    ):
        if position is None:
            index = self.addTab(tab, icon, description, data=data)
            if focus:
                self.setCurrentIndex(self.count() - 1)
        else:
            self.insertTab(position, tab, data, icon, description)
            if focus:
                self.setCurrentIndex(position)

    def removeTab(self, index: int) -> None:
        widget = self.widget(index)
        super().removeTab(index)
        if widget in self._tab_data:
            del self._tab_data[widget]


if __name__ == "__main__":
    import sys

    from PyQt6.QtWidgets import QApplication, QWidget

    app = QApplication(sys.argv)

    tab_widget = DataTabWidget(str)
    tab_widget.setMovable(True)
    tab1 = QWidget()
    tab2 = QWidget()

    # Example of adding tabs with and without icons
    tab_widget.addTab(tab1, description="Tab &1", data="Data for Tab 1")
    tab_widget.addTab(tab2, description="Tab &2", data="Data for Tab 2")  # No icon

    # Connect tab change signal to a function to display current tab data
    def show_current_tab_data(index) -> None:
        data = tab_widget.getCurrentTabData()
        tab_widget.setToolTip(f"Data for current tab: {data}")

    tab_widget.currentChanged.connect(show_current_tab_data)

    tab_widget.show()
    sys.exit(app.exec())



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import sys

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import (
    QApplication,
    QDialog,
    QDialogButtonBox,
    QLabel,
    QScrollArea,
    QVBoxLayout,
    QWidget,
)

from bitcoin_safe.html import link


class LicenseDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle(self.tr("License Info"))
        self.setModal(True)
        self.initUI()

    def initUI(self):
        layout = QVBoxLayout()

        # Create a QWidget to hold the content
        content_widget = QWidget()
        content_layout = QVBoxLayout(content_widget)

        # Adding QLabel
        label = QLabel(
            """                       
<p><b>Bitcoin-Safe</b></p>                       

<p>Bitcoin-Safe: A bitcoin wallet for the entire family.<br>
Copyright (C) 2024  Andreas Griffin</p>

<p>This program is free software: you can redistribute it and/or modify<br>
it under the terms of version 3 of the GNU General Public License as<br>
published by the Free Software Foundation.</p>

<p>This program is distributed in the hope that it will be useful,<br>
but WITHOUT ANY WARRANTY; without even the implied warranty of<br>
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the<br>
GNU General Public License for more details.</p>

<p>You should have received a copy of the GNU General Public License<br>
along with this program.  If not, see {link}.</p>
                       
<p>THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.          </p>             
                       
<p><b>Included libraries and software components</b></p>                       
<p><b>UR</b></p>                       

    <p>This software uses the 'ur' library, copyrighted by Foundation Devices, Inc. &copy; 2020. The 'ur' library is provided under the BSD-2-Clause Plus Patent License. The terms of this license permit redistribution and use in source and binary forms, with or without modification, subject to the following conditions:</p>
    <ol>
        <li>Redistributions of the source code must retain the above copyright notice, this list of conditions, and the following disclaimer.</li>
        <li>Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the following disclaimer in the documentation and/or other materials provided with the distribution.</li>
    </ol>
    <p>THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES.</p>

<p><b>Electrum</b></p>                                       


<p>Electrum - lightweight Bitcoin client<br>
Copyright (C) 2023 The Electrum Developers</p>    

<p>Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:</p>    

<p>The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.</p>    

<p>THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.</p>    
                    
                       """.format(
                link=link("https://www.gnu.org/licenses/gpl-3.0.html")
            ),
            self,
        )
        label.setOpenExternalLinks(True)  # Allows opening links
        label.setTextFormat(Qt.TextFormat.RichText)  # Set text as rich text
        label.setWordWrap(True)
        content_layout.addWidget(label)

        # Create a QScrollArea
        scroll_area = QScrollArea(self)
        scroll_area.setWidgetResizable(True)  # Important to make the scroll area adapt to the content
        scroll_area.setWidget(content_widget)
        layout.addWidget(scroll_area)

        # Adding QDialogButtonBox
        button_box = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok)
        button_box.accepted.connect(self.accept)
        layout.addWidget(button_box)

        self.setLayout(layout)


if __name__ == "__main__":
    app = QApplication(sys.argv)
    dialog = LicenseDialog()
    dialog.show()
    sys.exit(app.exec())



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
import sys
import threading
from collections import deque
from threading import Lock
from types import TracebackType
from typing import Any, Callable, NamedTuple, Optional, Tuple

from PyQt6.QtCore import QObject, QThread, pyqtSignal, pyqtSlot

from bitcoin_safe.execute_config import ENABLE_THREADING

logger = logging.getLogger(__name__)


class Task(NamedTuple):
    """Structure for task details."""

    do: Callable[..., Any]
    cb_success: Callable[[Any], None]
    cb_done: Callable[[Any], None]
    cb_error: Callable[[Tuple[Any, ...]], None]
    cancel: Optional[Callable[[], None]] = None


class Worker(QObject):
    """Worker object to perform tasks in a separate thread."""

    finished: pyqtSignal = pyqtSignal(object, object, object)  # Result, cb_done, cb_success/error
    error: pyqtSignal = pyqtSignal(object, object)

    def __init__(self, task: Task) -> None:
        super().__init__()
        self.task: Task = task
        self.thread_name = str(self.task.do)

    @pyqtSlot()
    def run_task(self) -> None:
        """Executes the provided task and emits signals based on the outcome."""
        threading.current_thread().name = self.thread_name
        try:
            logger.debug(f"Task started: {self.task.do}")
            result: Any = self.task.do()
            logger.debug(f"Task finished: {self.task.do}")
            self.finished.emit(result, self.task.cb_done, self.task.cb_success)
        except Exception:
            logger.exception(f"Task raised an exception: {self.task.do}")
            self.error.emit(sys.exc_info(), self.task.cb_error)
        finally:
            if callable(self.task.cancel):
                logger.debug(f"Task cancellation callback called: {self.task.do}")
                self.task.cancel()


class TaskThread(QThread):
    """Manages execution of tasks in separate threads."""

    signal_stop_threat = pyqtSignal(str)

    def __init__(self, enable_threading: bool = ENABLE_THREADING) -> None:
        super().__init__()
        self.worker: Optional[Worker] = (
            None  # Type hint adjusted because it will be immediately initialized in add_and_start
        )
        self.enable_threading = enable_threading
        self._thread_name: str | None = None

    @property
    def thread_name(self) -> str | None:
        return self._thread_name

    @thread_name.setter
    def thread_name(self, value: str | None):
        logger.debug(f"setting thread_name of {self} to {value}")
        self._thread_name = value

    def __str__(self) -> str:
        return str(self.thread_name)

    def add_and_start(
        self,
        do: Callable[..., Any],
        on_success: Callable[[Any], None],
        on_done: Callable[[Any], None],
        on_error: Callable[[Tuple[Any, ...]], None],
        cancel: Optional[Callable[[], None]] = None,
    ) -> "TaskThread":

        # if not self.enable_threading:
        #     NoThread().add_and_start(do, on_success, on_done, on_error)
        #     return self

        self.cancelled = False

        task: Task = Task(do, on_success, on_done, on_error, cancel)
        self.worker = Worker(task)
        if self.enable_threading:
            self.worker.moveToThread(self)
        self.worker.finished.connect(self.on_done)
        self.worker.error.connect(self.on_error)
        self.started.connect(self.worker.run_task)

        self.thread_name = self.worker.thread_name
        logger.debug(f"Starting new thread {self.thread_name}")

        self.start()

        return self

    def on_error(
        self,
        error_info: tuple[type[BaseException], BaseException, TracebackType],
        cb_error: Callable[[Any], None],
    ):
        if self.worker:
            cb_error(error_info)
        self.my_quit()

    @pyqtSlot(object, object, object)
    def on_done(self, result: Any, cb_done: Callable[[Any], None], cb_result: Callable[[Any], None]) -> None:
        """Handle task completion."""
        if not self.cancelled:
            logger.debug(f"Thread done: {self.thread_name}.")
            cb_done(result)
            cb_result(result)
        self.my_quit()

    def stop(self) -> None:
        """Stops the thread and any associated task cancellation if defined."""
        logger.debug(f"Stopping {self.thread_name}.")
        if self.worker and self.worker.task.cancel:
            self.cancelled = True
            self.worker.task.cancel()
        self.my_quit()
        logger.debug(f"Stopped {self.thread_name}.")

    def my_quit(self):
        try:
            self.cancelled = True
            self.quit()
            self.wait()
            self.signal_stop_threat.emit(self.thread_name)
        except:
            logger.error(f"An error during the shutdown of {self.thread_name}")


class NoThread:
    "This is great for debugging purposes"

    def __init__(self, *args, **kwargs):
        pass

    def add_and_start(
        self,
        task,
        on_success: Optional[Callable] = None,
        on_done: Optional[Callable] = None,
        on_error: Optional[Callable] = None,
    ):
        result = None
        try:
            if task:
                result = task()
            if on_success:
                on_success(result)
        except Exception:
            if on_error:
                on_error(sys.exc_info())
        if on_done:
            on_done(result)


class ThreadingManager:
    def __init__(
        self, threading_parent: "ThreadingManager" = None, threading_manager_name=None, **kwargs  # type: ignore
    ) -> None:
        super().__init__(**kwargs)
        self._taskthreads: deque[TaskThread] = deque()
        self.threading_manager_children: deque[ThreadingManager] = deque()
        self.lock = Lock()
        self.threading_parent = threading_parent
        self.threading_manager_name = (
            threading_manager_name if threading_manager_name else self.__class__.__name__
        )

        if threading_parent:
            threading_parent.threading_manager_children.append(self)

    def append_thread(self, thread: TaskThread):
        with self.lock:
            self._taskthreads.append(thread)
            thread.signal_stop_threat.connect(self.remove_thread)
            logger.debug(
                f"Appended thread {thread.thread_name} to {self.threading_manager_name}, Number of threads = {len(self._taskthreads)} {[str(thread) for thread in  self._taskthreads]}"
            )

    def remove_thread(self, thread_name: str | None):
        with self.lock:
            for thread in list(self._taskthreads):
                # if not thread.thread_name:
                #     # remove empty threads
                #     # (unclear why thread_name is set to None when threads are done)
                #     self.taskthreads.remove(thread)
                if thread.thread_name == thread_name:
                    self._taskthreads.remove(thread)
            logger.debug(
                f"Removed thread {thread_name} from {self.threading_manager_name}, Number of threads = {len(self._taskthreads)} {[str(thread) for thread in  self._taskthreads]}"
            )

    def stop_and_wait_all(self):
        while self.threading_manager_children:
            child = self.threading_manager_children.pop()
            child.end_threading_manager()

        # Wait for all threads to finish
        while self._taskthreads:
            taskthread = self._taskthreads.pop()
            logger.debug(f"stop taskthreads {taskthread}")
            taskthread.stop()

    def end_threading_manager(self):
        logger.debug(f"end_threading_manager of {self}")
        self.stop_and_wait_all()

        if self.threading_parent and self in self.threading_parent.threading_manager_children:
            self.threading_parent.threading_manager_children.remove(self)



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import io
import logging
import os
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any, List, Optional

from bitcoin_qr_tools.qr_generator import QRGenerator
from bitcoin_usb.address_types import DescriptorInfo
from PIL.Image import Image as PilImage
from reportlab.lib import colors
from reportlab.lib.enums import TA_CENTER
from reportlab.lib.pagesizes import letter
from reportlab.lib.styles import ParagraphStyle, getSampleStyleSheet
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.cidfonts import UnicodeCIDFont
from reportlab.platypus import (
    Image,
    PageBreak,
    Paragraph,
    SimpleDocTemplate,
    Spacer,
    Table,
    TableStyle,
)

from bitcoin_safe.i18n import translate

from .gui.qt.util import qicon_to_pil, read_QIcon, xdg_open_file
from .wallet import Wallet

logger = logging.getLogger(__name__)


TEXT_24_WORDS = translate("pdf", "12 or 24")


def pilimage_to_reportlab(pilimage: PilImage, width=200, height=200) -> Image:
    buffer = io.BytesIO()
    pilimage.save(buffer, format="PNG")
    buffer.seek(0)
    return Image(buffer, width=width, height=height)


# Define an enum for font types
class FontType(Enum):
    CID = "cid"
    BUILTIN = "builtin"


@dataclass
class FontInfo:
    font_name: str
    font_type: FontType
    supported_lang_code: str


def register_font(lang_code: str) -> FontInfo:
    """
    Registers a font for the given language code and returns a FontInfo object that contains
    the font details, font type, and supported language.

    If the language is not fully supported, it falls back to 'en_US' and uses Helvetica.

    :param lang_code: The language code to register the font for (e.g., 'zh_CN', 'ru_RU').
    :return: A FontInfo object with the font details and the supported language.
    """
    # Mapping language codes to FontInfo instances
    FONT_MAP: dict[str, FontInfo] = {
        # CID Fonts
        "zh_CN": FontInfo("STSong-Light", FontType.CID, "zh_CN"),  # Simplified Chinese
        "zh_TW": FontInfo("MSung-Light", FontType.CID, "zh_TW"),  # Traditional Chinese (Taiwan)
        "zh_HK": FontInfo("MHei-Medium", FontType.CID, "zh_HK"),  # Traditional Chinese (Hong Kong)
        "ja_JP": FontInfo("HeiseiMin-W3", FontType.CID, "ja_JP"),  # Japanese
        "ko_KR": FontInfo("HYGoThic-Medium", FontType.CID, "ko_KR"),  # Korean
        # Built-in Fonts (Latin-based)
        "es_ES": FontInfo("Helvetica", FontType.BUILTIN, "es_ES"),  # Spanish
        "fr_FR": FontInfo("Helvetica", FontType.BUILTIN, "fr_FR"),  # French
        "en_US": FontInfo("Helvetica", FontType.BUILTIN, "en_US"),  # English (US)
        "en_GB": FontInfo("Helvetica", FontType.BUILTIN, "en_GB"),  # English (UK)
        "pt_PT": FontInfo("Helvetica", FontType.BUILTIN, "pt_PT"),  # Portuguese (Portugal)
        "pt_BR": FontInfo("Helvetica", FontType.BUILTIN, "pt_BR"),  # Portuguese (Brazil)
        "it_IT": FontInfo("Helvetica", FontType.BUILTIN, "it_IT"),  # Italian
        "de_DE": FontInfo("Helvetica", FontType.BUILTIN, "de_DE"),  # German
        "nl_NL": FontInfo("Helvetica", FontType.BUILTIN, "nl_NL"),  # Dutch (Netherlands)
        "nl_BE": FontInfo("Helvetica", FontType.BUILTIN, "nl_BE"),  # Dutch (Belgium)
        "sv_SE": FontInfo("Helvetica", FontType.BUILTIN, "sv_SE"),  # Swedish
        "da_DK": FontInfo("Helvetica", FontType.BUILTIN, "da_DK"),  # Danish
        "no_NO": FontInfo("Helvetica", FontType.BUILTIN, "no_NO"),  # Norwegian
        "fi_FI": FontInfo("Helvetica", FontType.BUILTIN, "fi_FI"),  # Finnish
        "is_IS": FontInfo("Helvetica", FontType.BUILTIN, "is_IS"),  # Icelandic
        "pl_PL": FontInfo("Helvetica", FontType.BUILTIN, "pl_PL"),  # Polish
        "cs_CZ": FontInfo("Helvetica", FontType.BUILTIN, "cs_CZ"),  # Czech
        "sk_SK": FontInfo("Helvetica", FontType.BUILTIN, "sk_SK"),  # Slovak
        "sl_SI": FontInfo("Helvetica", FontType.BUILTIN, "sl_SI"),  # Slovenian
        "hu_HU": FontInfo("Helvetica", FontType.BUILTIN, "hu_HU"),  # Hungarian
        "ro_RO": FontInfo("Helvetica", FontType.BUILTIN, "ro_RO"),  # Romanian
        "hr_HR": FontInfo("Helvetica", FontType.BUILTIN, "hr_HR"),  # Croatian
        "sr_RS": FontInfo("Helvetica", FontType.BUILTIN, "sr_RS"),  # Serbian (Latin)
        "bs_BA": FontInfo("Helvetica", FontType.BUILTIN, "bs_BA"),  # Bosnian
        "mk_MK": FontInfo("Helvetica", FontType.BUILTIN, "mk_MK"),  # Macedonian (Latin)
        "mt_MT": FontInfo("Helvetica", FontType.BUILTIN, "mt_MT"),  # Maltese
        "gl_ES": FontInfo("Helvetica", FontType.BUILTIN, "gl_ES"),  # Galician
        "ca_ES": FontInfo("Helvetica", FontType.BUILTIN, "ca_ES"),  # Catalan
        "eu_ES": FontInfo("Helvetica", FontType.BUILTIN, "eu_ES"),  # Basque
        "lv_LV": FontInfo("Helvetica", FontType.BUILTIN, "lv_LV"),  # Latvian
        "lt_LT": FontInfo("Helvetica", FontType.BUILTIN, "lt_LT"),  # Lithuanian
        "et_EE": FontInfo("Helvetica", FontType.BUILTIN, "et_EE"),  # Estonian
        "af_ZA": FontInfo("Helvetica", FontType.BUILTIN, "af_ZA"),  # Afrikaans
        "vi_VN": FontInfo("Helvetica", FontType.BUILTIN, "vi_VN"),  # Vietnamese (Latin-based characters)
        "ms_MY": FontInfo("Helvetica", FontType.BUILTIN, "ms_MY"),  # Malay
        "id_ID": FontInfo("Helvetica", FontType.BUILTIN, "id_ID"),  # Indonesian
        "en_US": FontInfo("Helvetica", FontType.BUILTIN, "en_US"),  # English
    }

    if lang_code in FONT_MAP:
        font_info: FontInfo = FONT_MAP[lang_code]

        if font_info.font_type == FontType.CID:
            # Register the CID font
            pdfmetrics.registerFont(UnicodeCIDFont(font_info.font_name))
            print(f"Using built-in CID font: {font_info.font_name} for language code: {lang_code}")
        elif font_info.font_type == FontType.BUILTIN:
            # No registration needed for built-in fonts like Helvetica
            print(f"Using built-in font: {font_info.font_name} for language code: {lang_code}")

        return font_info

    else:
        print(f"No font found for language code: {lang_code}, returning en_US")
        return FontInfo("Helvetica", FontType.BUILTIN, "en_US")  # Default fallback to en_US


white_space = '<font color="white"> - </font>'


class BitcoinWalletRecoveryPDF:

    def __init__(self, lang_code: str) -> None:
        font_info = register_font(lang_code=lang_code)
        self.font_name = font_info.font_name
        self.no_translate = font_info.supported_lang_code == "en_US"

        styles = getSampleStyleSheet()
        self.style_paragraph = ParagraphStyle(
            name="Centered",
            fontName=self.font_name,
            parent=styles["BodyText"],
            alignment=TA_CENTER,
        )
        self.style_paragraph_left = ParagraphStyle(
            name="LEFT",
            fontName=self.font_name,
            parent=styles["BodyText"],
        )
        self.style_heading = ParagraphStyle(
            "centered_heading",
            fontName=self.font_name,
            parent=styles["Heading1"],
            alignment=TA_CENTER,
        )
        self.style_text = ParagraphStyle(
            name="normal",
            fontName=self.font_name,
        )
        self.elements: List[Any] = []

    @property
    def TEXT_24_WORDS(self):
        return translate("pdf", "12 or 24", no_translate=self.no_translate)

    @staticmethod
    def create_table(columns: List[Any], col_widths: List[int]) -> Table:
        # Validate input and create data for the table
        max_rows = max([len(col) for col in columns])
        data = []
        for i in range(max_rows):
            row = [col[i] if i < len(col) else "" for col in columns]
            data.append(row)

        # Create a Table with data and specify column widths
        table = Table(data, colWidths=col_widths)

        # Apply TableStyle to make the borders invisible
        style = TableStyle(
            [
                ("BOX", (0, 0), (-1, -1), 0, colors.white),  # Outer border
                ("INNERGRID", (0, 0), (-1, -1), 0, colors.white),  # Inner grid
                ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),  # Vertical alignment for all cells
                ("ALIGN", (0, 0), (-1, -1), "CENTER"),  # Vertical alignment for all cells
                # ('VALIGN', (0, 0), (0, 0), 'TOP'),                  # Vertical alignment for first cell
                # ('VALIGN', (1, 0), (1, 0), 'BOTTOM')                # Vertical alignment for second cell
            ]
        )

        table.setStyle(style)

        return table

    def add_page_break(self) -> None:
        self.elements.append(PageBreak())  # Add a page break between documents if needed

    def _seed_part(
        self,
        seed: Optional[str],
        keystore_description: str,
        keystore_fingerprint: str,
        keystore_label: str,
        num_signers: int,
    ) -> None:
        self.elements.append(Spacer(1, 5))
        # Additional subtitle
        if num_signers == 1:
            instructions1 = Paragraph(
                translate(
                    "pdf",
                    """1. Glue or tape the 'Recovery sheet' ({number} words) over the table below<br/>
                2. Fold this  paper at the line below <br/>
                3. Put this paper in a secure location, where only you have access<br/>
                4. You can put the hardware signer either a) together with the paper seed backup, or b)   in another secure  location (if available)   
                """,
                    no_translate=self.no_translate,
                ).format(number=self.TEXT_24_WORDS),
                self.style_paragraph_left,
            )
        else:
            instructions1 = Paragraph(
                translate(
                    "pdf",
                    """1. Glue or tape the 'Recovery sheet' ({number} words) over the table below<br/>
                2. Fold this  paper at the line below <br/>
                3. Put each paper in a different secure location, where only you have access<br/>
                4. You can put the hardware signers either a) together with the corresponding paper seed backup, or b)   each  in yet another secure  location (if available)   
                """,
                    no_translate=self.no_translate,
                ).format(number=self.TEXT_24_WORDS),
                self.style_paragraph_left,
            )

        # No photography icon
        icon = read_QIcon("no-typing-icon.svg")
        icon2 = read_QIcon("no-photography-icon.svg")
        reportlab_icon = pilimage_to_reportlab(qicon_to_pil(icon), width=50, height=50)
        reportlab_icon2 = pilimage_to_reportlab(qicon_to_pil(icon2), width=50, height=50)

        self.elements.append(
            self.create_table(
                [[reportlab_icon], [instructions1], [reportlab_icon2]],
                [60, 400, 60],
            )
        )

        self.elements.append(Spacer(1, 5))

        # Table title
        table_title = translate(
            "pdf",
            "Secret seed words for a hardware signer: Never type into a computer. Never make a picture.",
            no_translate=self.no_translate,
        )
        seed_placeholder = "___________________"

        # split seed words if available
        if seed:
            seed_items = seed.split(" ")
            seed_items = seed_items + [seed_placeholder for i in range(24 - len(seed_items))]
        else:
            seed_items = [seed_placeholder for i in range(24)]

        # 24 words placeholder in three columns
        data = [[table_title, "", ""]]  # First row is the title
        for i in range(8):
            data.append([f"{i + j+1} {seed_items[i+j]}" for j in [0, 8, 16]])

        table = Table(data)

        # Table styling with border and title formatting
        table_style = TableStyle(
            [
                ("BOX", (0, 0), (-1, -1), 1, colors.black),
                ("ALIGN", (0, 0), (-1, -1), "CENTER"),
                ("SPAN", (0, 0), (2, 0)),  # Merging the cells of the title row
                (
                    "BACKGROUND",
                    (0, 0),
                    (2, 0),
                    colors.grey,
                ),  # Background color for the title
                (
                    "TEXTCOLOR",
                    (0, 0),
                    (2, 0),
                    colors.whitesmoke,
                ),  # Text color for the title
                ("FONTNAME", (0, 0), (2, 0), self.font_name),  # Font for the title
                ("ALIGN", (0, 0), (-1, -1), "LEFT"),  # Vertical alignment for all cells
                ("LINEBELOW", (0, -1), (-1, -1), 1, colors.black),  # Add the closing horizontal line below
            ],
            fontName=self.font_name,
        )

        table.setStyle(table_style)
        table.hAlign = "CENTER"
        self.elements.append(table)
        self.elements.append(Spacer(1, 1))

        description_text = Paragraph(
            translate(
                "pdf",
                "{keystore_label} ({keystore_fingerprint}): {keystore_description}<br/><br/>Instructions for the heirs:",
                no_translate=self.no_translate,
            ).format(
                keystore_description=keystore_description.replace("\n", "<br/>"),
                keystore_fingerprint=keystore_fingerprint,
                keystore_label=keystore_label,
            ),
            self.style_paragraph_left,
        )

        self.elements.append(
            self.create_table(
                [[reportlab_icon2], [description_text], [reportlab_icon]],
                [60, 400, 60],
            )
        )

    def _descriptor_part(
        self,
        wallet_descriptor_string: str,
        threshold: int,
    ) -> None:
        qr_image = pilimage_to_reportlab(
            QRGenerator.create_qr_PILimage(wallet_descriptor_string), width=200, height=200
        )
        if threshold > 1:
            desc_str = Paragraph(
                translate(
                    "pdf",
                    "The wallet descriptor (QR Code) <br/><br/>{wallet_descriptor_string}<br/><br/> allows you to create a watch-only wallet to see your balance. To spent from it you need {threshold} Seeds and the wallet descriptor.",
                    no_translate=self.no_translate,
                ).format(threshold=threshold, wallet_descriptor_string=wallet_descriptor_string),
                self.style_paragraph,
            )
        else:
            desc_str = Paragraph(
                translate(
                    "pdf",
                    "The wallet descriptor (QR Code) <br/><br/>{wallet_descriptor_string}<br/><br/> allows you to create a watch-only wallet to see your balance.  To spent from it you need the secret {number} words (Seed).",
                    no_translate=self.no_translate,
                ).format(number=self.TEXT_24_WORDS, wallet_descriptor_string=wallet_descriptor_string),
                self.style_paragraph,
            )
        self.elements.append(self.create_table([[qr_image], [desc_str]], [250, 300]))

    def create_pdf(
        self,
        title: str,
        wallet_descriptor_string: str,
        keystore_description: str,
        keystore_label: str,
        keystore_xpub: str,
        keystore_key_origin: str,
        keystore_fingerprint: str,
        threshold: int,
        seed: Optional[str] = None,
        num_signers: int = 1,
    ) -> None:
        self.elements.append(Paragraph(title, style=self.style_heading))

        # Small subtitle
        self.elements.append(
            Paragraph(
                translate("pdf", "Created with", no_translate=self.no_translate)
                + f" Bitcoin Safe: {white_space*5} https://github.com/andreasgriffin/bitcoin-safe",
                self.style_paragraph,
            )
        )
        self.elements.append(Paragraph(f"", self.style_paragraph))

        self._seed_part(
            seed,
            num_signers=num_signers,
            keystore_label=keystore_label,
            keystore_fingerprint=keystore_fingerprint,
            keystore_description=keystore_description,
        )

        self.elements.append(Spacer(1, 15))

        # Add a horizontal line as an element
        line = Paragraph(
            "________________________________________________________________________________",
            self.style_paragraph,
        )
        # line.keepWithNext = True  # Ensure line and text stay together on the same page
        self.elements.append(line)
        # Add text at the line as an element
        text = (white_space * 5).join(
            [translate("pdf", "Please fold here!", no_translate=self.no_translate)] * 5
        )
        text_paragraph = Paragraph(text, style=self.style_text)
        # text_paragraph.spaceBefore = -10  # Adjust the space before the text if needed
        self.elements.append(text_paragraph)

        self._descriptor_part(wallet_descriptor_string, threshold)

        keystore_info_text = Paragraph(
            translate(
                "pdf",
                "{keystore_label}: Fingerprint: {keystore_fingerprint}, Key origin: {keystore_key_origin}, {keystore_xpub}",
                no_translate=self.no_translate,
            ).format(
                keystore_label=keystore_label,
                keystore_fingerprint=keystore_fingerprint,
                keystore_key_origin=keystore_key_origin,
                keystore_xpub=keystore_xpub,
            ),
            self.style_paragraph_left,
        )
        self.elements.append(keystore_info_text)

    def save_pdf(self, filename: str) -> None:

        # Adjust these values to set your desired margins (values are in points; 72 points = 1 inch)
        LEFT_MARGIN = 36  # 0.5 inch
        RIGHT_MARGIN = 36  # 0.5 inch
        TOP_MARGIN = 36  # 0.5 inch
        BOTTOM_MARGIN = 36  # 0.5 inch

        document = SimpleDocTemplate(
            filename,
            pagesize=letter,
            leftMargin=LEFT_MARGIN,
            rightMargin=RIGHT_MARGIN,
            topMargin=TOP_MARGIN,
            bottomMargin=BOTTOM_MARGIN,
        )
        document.build(self.elements)

    def open_pdf(self, filename: str) -> None:
        if os.path.exists(filename):
            xdg_open_file(Path(filename))
        else:
            logger.info("File not found!")


def make_and_open_pdf(wallet: Wallet, lang_code: str) -> None:
    info = DescriptorInfo.from_str(wallet.multipath_descriptor.as_string())
    pdf_recovery = BitcoinWalletRecoveryPDF(lang_code=lang_code)

    for i, keystore in enumerate(wallet.keystores):
        title = (
            translate(
                "pdf",
                '{i}. Seed backup of a {threshold} of {m} Multi-Sig Wallet: "{id}"',
                no_translate=pdf_recovery.no_translate,
            ).format(i=i + 1, threshold=info.threshold, m=len(wallet.keystores), id=wallet.id)
            if len(wallet.keystores) > 1
            else f"{wallet.id }"
        )
        pdf_recovery.create_pdf(
            title=title,
            wallet_descriptor_string=wallet.multipath_descriptor.as_string(),
            threshold=info.threshold,
            seed=keystore.mnemonic,
            num_signers=len(wallet.keystores),
            keystore_xpub=keystore.xpub,
            keystore_description=keystore.description,
            keystore_fingerprint=keystore.fingerprint,
            keystore_key_origin=keystore.key_origin,
            keystore_label=keystore.label,
        )
        pdf_recovery.add_page_break()
    temp_file = os.path.join(
        Path.home(), translate("pdf", "Seed backup of {id}").format(id=wallet.id) + ".pdf"
    )
    pdf_recovery.save_pdf(temp_file)
    pdf_recovery.open_pdf(temp_file)


# # Example Usage
# wallet_name = "My Wallet Name"
# wallet_descriptor_qr_code = Image("../bitcoin_safe/gui/icons/qrcode.png")  # Replace with path to your QR code image
# wallet_descriptor_string = "Your wallet descriptor string here"
# wallet_description = "Your wallet description here"

# pdf_creator = BitcoinWalletRecoveryPDF(wallet_name, wallet_descriptor_qr_code, wallet_descriptor_string, wallet_description)
# pdf_creator.create_pdf()
# pdf_filename = "wallet_recovery.pdf"
# pdf_creator.save_pdf(pdf_filename)
# pdf_creator.open_pdf(pdf_filename)



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
from dataclasses import dataclass

from bitcoin_safe.signals import SignalsMin
from bitcoin_safe.threading_manager import ThreadingManager

logger = logging.getLogger(__name__)

import os
import sys
from math import ceil
from typing import Callable, List, Optional, Union

from PyQt6.QtCore import QEvent, QPoint, QRect, QRectF, QSize, Qt, pyqtSignal
from PyQt6.QtGui import (
    QBrush,
    QColor,
    QFont,
    QFontMetrics,
    QMouseEvent,
    QPainter,
    QPaintEvent,
    QPen,
    QPixmap,
    QPolygon,
    QResizeEvent,
    QTextOption,
)
from PyQt6.QtWidgets import (
    QApplication,
    QDialogButtonBox,
    QPushButton,
    QSizePolicy,
    QStackedWidget,
    QTextEdit,
    QToolTip,
    QVBoxLayout,
    QWidget,
)

from bitcoin_safe.gui.qt.util import create_button_box


def height_of_str(text, widget: QWidget, max_width: float) -> float:
    font_metrics = QFontMetrics(widget.font())
    rect = font_metrics.boundingRect(QRect(0, 0, int(max_width * 0.95), 1000), Qt.TextFlag.TextWordWrap, text)
    return rect.height() * 1.2


class StepProgressBar(QWidget):
    # Define a new signal that emits the number of the index clicked
    signal_index_clicked = pyqtSignal(int)

    def __init__(
        self,
        number_of_steps: int,
        current_index: int = 0,
        parent=None,
        mark_current_index_as_completed=False,
        clickable=True,
        use_checkmark_icon=True,
        circle_sizes: List[int] | None = None,
    ) -> None:
        super().__init__(parent)
        self.current_index = current_index
        self.clickable = clickable
        self.enumeration_alphabet: Optional[List[str]] = None
        self.cursor_set = False
        self.labels = [f"Step {i+1}" for i in range(number_of_steps)]  # Default labels
        self.set_circle_sizes(circle_sizes)
        self.tube_width = 4
        self.padding = 0
        self.label_distance = 5
        self.max_label_height = 0  # Initialize max label height
        self.use_checkmark_icon = use_checkmark_icon

        normalized_icon_path = os.path.normpath(
            os.path.join(os.path.dirname(__file__), "../icons/checkmark.svg")
        )
        self.checkmark_pixmap = QPixmap(normalized_icon_path).scaled(
            20, 20, Qt.AspectRatioMode.KeepAspectRatio, Qt.TransformationMode.SmoothTransformation
        )  # Load and scale your checkmark image
        self.mark_current_index_as_completed = mark_current_index_as_completed
        self.tooltips = [""] * number_of_steps  # Initialize tooltips as empty strings
        self.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)

    def set_circle_sizes(self, circle_sizes: List[int] | None = None) -> None:
        self.radius = max(circle_sizes) if circle_sizes else 20
        self.circle_sizes = (
            circle_sizes if circle_sizes else [self.radius for i in range(self.number_of_steps)]
        )
        if len(self.circle_sizes) < self.number_of_steps:
            self.circle_sizes = self.circle_sizes * ceil(self.number_of_steps / len(self.circle_sizes))

    def set_enumeration_alphabet(self, enumeration_alphabet: Optional[List[str]]) -> None:
        self.enumeration_alphabet = enumeration_alphabet

    def set_current_index(self, index: int) -> None:
        self.current_index = index
        self.update()  # Redraw the widget

    def recalculate_max_height(self) -> None:
        max_width = self.width() / (self.number_of_steps + 1)
        self.max_label_height = 0
        for label in self.labels:
            self.max_label_height = int(max(self.max_label_height, height_of_str(label, self, max_width)))
        self.updateGeometry()  # Notify the layout system that the widget's size hint has changed

    def resizeEvent(self, event: QResizeEvent | None) -> None:
        self.recalculate_max_height()
        super().resizeEvent(event)

    @property
    def number_of_steps(self) -> int:
        return len(self.labels)

    def set_labels(self, labels: List[str]) -> None:
        self.labels = labels
        self.set_circle_sizes()
        self.tooltips = [""] * self.number_of_steps

        self.recalculate_max_height()
        self.update()

    def sizeHint(self) -> QSize:
        total_height = int(
            self.radius * 2 + max(self.max_label_height, 20) + self.label_distance + self.padding * 2
        )
        return QSize(self.width(), total_height)

    def mousePressEvent(self, event: QMouseEvent | None) -> None:
        if not event:
            super().mousePressEvent(event)
            return

        for i in range(self.number_of_steps):
            # Define the rectangle area for each step
            if self._ellipse_rect(i).contains(event.position()):
                self.signal_index_clicked.emit(i)  # Emit the clicked step number
                break

        super().mousePressEvent(event)

    def paintEvent(self, event: QPaintEvent | None) -> None:
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        # Define pens, brushes, and colors
        completed_color = QColor("#7FBA00")
        current_color = QColor("#0078D4")
        inactive_color = QColor("#BFBFBF")
        text_color = QColor("#505050")
        bubble_color = QColor("#FFFFFF")

        QPen(current_color, self.tube_width)
        completed_pen = QPen(completed_color, self.tube_width)
        inactive_pen = QPen(inactive_color, self.tube_width)
        QBrush(current_color)
        completed_brush = QBrush(completed_color)
        inactive_brush = QBrush(inactive_color)

        step_width = self._step_width()

        # Create a QTextOption for word wrapping
        text_option = QTextOption()
        text_option.setWrapMode(QTextOption.WrapMode.WordWrap)
        text_option.setAlignment(Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignTop)

        for i in range(self.number_of_steps):
            ellipse_rect = self._ellipse_rect(i)
            center_point = ellipse_rect.center()

            is_past_step = i < self.current_index or (
                i == self.current_index and self.mark_current_index_as_completed
            )

            # Draw connecting line
            if i < self.number_of_steps - 1:
                next_color = completed_color if i < self.current_index else inactive_color
                painter.setPen(QPen(next_color, self.tube_width))
                painter.drawLine(
                    int(center_point.x()),
                    int(center_point.y()),
                    int(center_point.x() + step_width),
                    int(center_point.y()),
                )

            # Set pen and brush for circle
            if is_past_step:
                painter.setPen(completed_pen)
                painter.setBrush(completed_brush)
            elif i == self.current_index:
                color = completed_color if self.mark_current_index_as_completed else current_color
                painter.setPen(QPen(color, self.tube_width))
                painter.setBrush(QBrush(color))
            else:
                painter.setPen(inactive_pen)
                painter.setBrush(inactive_brush)

            # Draw circle
            painter.drawEllipse(ellipse_rect)

            # Draw checkmark icon or step number
            if self.use_checkmark_icon and (
                is_past_step or (i == self.current_index and self.mark_current_index_as_completed)
            ):
                icon_size = self.checkmark_pixmap.size()
                painter.drawPixmap(
                    int(center_point.x() - icon_size.width() / 2),
                    int(center_point.y() - icon_size.height() / 2),
                    self.checkmark_pixmap,
                )
            else:
                painter.setPen(QPen(bubble_color))
                painter.drawText(
                    ellipse_rect,
                    Qt.AlignmentFlag.AlignCenter,
                    self.format_step(i),
                )

            # Draw step text below circles
            painter.setPen(QPen(text_color))
            painter.setFont(QFont("Arial", 10))
            painter.drawText(self._label_rect(i), self.labels[i], text_option)

        painter.end()

    def format_step(self, i: int) -> str:
        return (
            str(i + 1)
            if self.enumeration_alphabet is None
            else self.enumeration_alphabet[i % len(self.enumeration_alphabet)]
        )

    def _step_width(self) -> float:
        return self.width() / (self.number_of_steps + 1)

    def _label_rect(self, i: int) -> QRectF:
        circle_y = self.radius + self.tube_width  # Position circles near the top
        step_width = self._step_width()
        return QRectF(
            step_width * (i + 1) - step_width / 2,
            circle_y + self.radius + self.padding + self.label_distance,
            step_width,
            self.max_label_height,
        )

    def _ellipse_rect(self, i: int) -> QRectF:
        circle_y = self.radius + self.tube_width  # Position circles near the top
        step_width = self._step_width()
        return QRectF(
            step_width * (i + 1) - self.circle_sizes[i],
            circle_y - self.circle_sizes[i],
            2 * self.circle_sizes[i],
            2 * self.circle_sizes[i],
        )

    def set_mark_current_step_as_completed(self, value: bool) -> None:
        self.mark_current_index_as_completed = value
        self.update()

    def set_step_tooltips(self, tooltips: List[str]) -> None:
        self.tooltips = tooltips + ["" for i in range(len(tooltips), self.number_of_steps)]

    def enterEvent(self, event: QEvent | None) -> None:
        self.setMouseTracking(True)  # Enable mouse tracking to receive mouse move events

    def leaveEvent(self, event: QEvent | None) -> None:
        self.setMouseTracking(False)  # Disable mouse tracking when the mouse leaves the widget
        QToolTip.hideText()  # Hide tooltip when the cursor is not above a step
        self.restore_cursor()

    def mouseMoveEvent(self, event: QMouseEvent | None) -> None:
        if not event:
            super().mouseMoveEvent(event)
            return

        in_circle = None
        for i in range(self.number_of_steps):
            if self._ellipse_rect(i).contains(event.position()):
                in_circle = i
                break

        if in_circle is None:
            QToolTip.hideText()  # Hide tooltip if the cursor is not above any step
            self.restore_cursor()
        else:
            self.set_cursor()
            QToolTip.showText(event.globalPosition().toPoint(), self.tooltips[in_circle], self)

        super().mouseMoveEvent(event)

    def restore_cursor(self) -> None:
        if self.clickable and self.cursor_set:
            self.cursor_set = False
            QApplication.restoreOverrideCursor()  # Restore to default cursor

    def set_cursor(self) -> None:
        if self.clickable and not self.cursor_set:
            self.cursor_set = True
            QApplication.setOverrideCursor(Qt.CursorShape.PointingHandCursor)


class HorizontalIndicator(QWidget):
    def __init__(self, number_of_steps: int, current_index: int, parent=None) -> None:
        super().__init__(parent)
        self.number_of_steps = number_of_steps
        self.current_index = current_index
        self.pen_width = 2
        self.triangle_size = 13 + self.pen_width  # Half-width of the triangle
        self.triangle_height = 1 * self.triangle_size
        self.setMinimumHeight(
            self.triangle_height + self.pen_width
        )  # Set minimum height to ensure there's enough space for the line and triangle

        self.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)

    def set_number_of_steps(self, number_of_steps: int) -> None:
        self.number_of_steps = number_of_steps
        self.update()

    def set_current_index(self, index: int) -> None:
        self.current_index = index
        self.update()  # Repaint the widget with the new step indicator

    def paintEvent(self, event: QPaintEvent | None) -> None:
        super().paintEvent(event)
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        # Set background color using a QPalette
        # palette = self.palette()
        # palette.setColor(self.backgroundRole(), QColor("#ff0000"))  # Replace with your desired color
        # self.setPalette(palette)
        # self.setAutoFillBackground(True)

        color = QColor("#BFBFBF")
        line_y = self.height() - self.pen_width / 2  # Adjust this as needed

        # Draw the gray horizontal line
        pen = QPen(color, self.pen_width)
        painter.setPen(pen)
        painter.drawLine(0, int(line_y), int(self.width()), int(line_y))

        # Draw the triangle indicator pointing up
        step_width = self.width() / (self.number_of_steps + 1)
        triangle_center = QPoint(int(step_width * (self.current_index + 1)), int(line_y))
        triangle = QPolygon(
            [
                triangle_center - QPoint(int(self.triangle_size), 0),
                triangle_center + QPoint(int(self.triangle_size), 0),
                triangle_center - QPoint(0, int(self.triangle_height)),  # Pointing up
            ]
        )
        brush = QBrush(color)
        painter.setBrush(brush)

        pen = QPen(color, 1)
        painter.setPen(pen)
        painter.drawPolygon(triangle)


class AutoResizingStackedWidget(QWidget):
    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self._layout = QVBoxLayout(self)
        self.setLayout(self._layout)  # Explicitly setting the layout
        self._layout.setContentsMargins(0, 0, 0, 0)  # Left, Top, Right, Bottom margins
        self.widgets: List[QWidget] = []
        self._currentIndex = -1

    def addWidget(self, widget: QWidget) -> None:
        self.insertWidget(len(self.widgets), widget)

    def insertWidget(self, index, widget: QWidget) -> None:
        if index < 0 or index > len(self.widgets):
            raise IndexError("Index out of bounds")
        widget.setVisible(False)
        self.widgets.insert(index, widget)
        self._layout.insertWidget(index, widget)
        if len(self.widgets) == 1:
            self.setCurrentIndex(0)

    def setCurrentIndex(self, index: int) -> None:
        if 0 <= index < len(self.widgets):
            if self._currentIndex != -1:
                self.widgets[self._currentIndex].setVisible(False)
            self.widgets[index].setVisible(True)

            self._currentIndex = index
            self.adjustSizeToCurrentWidget()

    def currentIndex(self) -> int:
        return self._currentIndex

    def adjustSizeToCurrentWidget(self) -> None:
        if self._currentIndex != -1:
            currentWidget = self.widgets[self._currentIndex]
            if currentWidget.sizeHint().height() != -1 and currentWidget.sizeHint().width() != -1:
                self.setMinimumHeight(currentWidget.sizeHint().height())
            self.setMaximumSize(currentWidget.maximumSize())

    def count(self) -> int:
        return len(self.widgets)

    def removeWidget(self, widget: QWidget) -> None:
        if widget in self.widgets:
            self.widgets.remove(widget)
            widget.setParent(None)  # type: ignore[call-overload]
            widget.deleteLater()  # This is important to fully remove the widget
            if self._currentIndex >= len(self.widgets):
                self.setCurrentIndex(len(self.widgets) - 1)

    def widget(self, index: int) -> Optional[QWidget]:
        if 0 <= index < len(self.widgets):
            return self.widgets[index]
        return None

    def currentWidget(self) -> Optional[QWidget]:
        if self._currentIndex != -1:
            return self.widgets[self._currentIndex]
        return None

    def indexOf(self, widget: QWidget) -> int:
        return self.widgets.index(widget) if widget in self.widgets else -1


class StepProgressContainer(ThreadingManager, QWidget):
    signal_set_current_widget = pyqtSignal(QWidget)
    signal_widget_focus = pyqtSignal(QWidget)
    signal_widget_unfocus = pyqtSignal(QWidget)

    def __init__(
        self,
        step_labels: List[str],
        signals_min: SignalsMin,
        current_index: int = 0,
        collapsible_current_active=False,
        clickable=True,
        use_checkmark_icon=True,
        parent=None,
        sub_indices: List[int] | None = None,
        use_resizing_stacked_widget=True,
        threading_parent: ThreadingManager | None = None,
    ) -> None:
        super().__init__(parent=parent, threading_parent=threading_parent)  # type: ignore
        self.signals_min = signals_min
        self.threading_parent = threading_parent
        self.step_bar = StepProgressBar(
            len(step_labels),
            current_index=current_index,
            clickable=clickable,
            use_checkmark_icon=use_checkmark_icon,
            circle_sizes=(
                None
                if sub_indices is None
                else [12 if i in sub_indices else 20 for i in range(len(step_labels))]
            ),
        )

        self.horizontal_indicator = HorizontalIndicator(len(step_labels), current_index)
        self.stacked_widget: Union[AutoResizingStackedWidget, QStackedWidget] = (
            AutoResizingStackedWidget() if use_resizing_stacked_widget else QStackedWidget()
        )
        self.collapsible_current_active = collapsible_current_active
        self.clickable = clickable

        self.set_labels(step_labels)

        self._layout = QVBoxLayout(self)
        self._layout.addWidget(self.step_bar)
        self._layout.addWidget(self.horizontal_indicator)
        self._layout.setSpacing(0)  # This sets the spacing between items in the layout to zero
        self._layout.addSpacing(5)
        self._layout.setContentsMargins(0, 0, 0, 0)

        self._layout.addWidget(self.stacked_widget)

        self.set_current_index(current_index)

        self.step_bar.signal_index_clicked.connect(self.on_click)

    def set_labels(self, labels: List[str]) -> None:
        self.step_bar.set_labels(labels=labels)
        self.horizontal_indicator.set_number_of_steps(len(labels))

        # reset widgets
        while self.stacked_widget.count() > len(labels):
            widget = self.stacked_widget.widget(0)
            if widget:
                self.stacked_widget.removeWidget(widget)
        for i in range(len(labels) - self.stacked_widget.count()):
            custom_widget = QWidget()
            self.stacked_widget.addWidget(custom_widget)

    def set_sub_indices(self, sub_indices: List[int] | None = None) -> None:
        self.step_bar.set_circle_sizes(
            None
            if sub_indices is None
            else [12 if i in sub_indices else 20 for i in range(len(self.step_bar.labels))]
        )

    def current_index(self) -> int:
        return self.step_bar.current_index

    def current_highlighted_index(self) -> int:
        return self.stacked_widget.currentIndex()

    def count(self) -> int:
        return self.step_bar.number_of_steps

    def on_click(self, index: int) -> None:
        if not self.clickable:
            return

        currently_visible_step = (
            self.stacked_widget.currentIndex() if self.stacked_widget.isVisible() else None
        )

        if currently_visible_step is None:
            if self.collapsible_current_active:
                self.set_stacked_widget_visible(True)
        else:
            if currently_visible_step == index:
                # hide the stacked widget
                if self.collapsible_current_active:
                    self.set_stacked_widget_visible(False)

        self.set_focus(index)

    def set_focus(self, index: int) -> None:
        old_index = self.stacked_widget.currentIndex()

        if old_index == index:
            return

        self.stacked_widget.setCurrentIndex(index)
        self.horizontal_indicator.set_current_index(index)

        self.signal_widget_unfocus.emit(self.stacked_widget.widget(old_index))
        self.signal_widget_focus.emit(self.stacked_widget.widget(index))

    def set_stacked_widget_visible(self, is_visible: bool) -> None:
        self.horizontal_indicator.setVisible(is_visible)
        self.stacked_widget.setVisible(is_visible)

    def set_current_index(self, index: int) -> None:
        old_index = self.current_index()

        if old_index == index:
            return

        self.step_bar.set_current_index(index)
        self.horizontal_indicator.set_current_index(index)
        self.stacked_widget.setCurrentIndex(index)

        # this order is important:
        self.signal_widget_unfocus.emit(self.stacked_widget.widget(old_index))
        self.signal_widget_focus.emit(self.stacked_widget.widget(index))
        # in the set_current_widget the callback functions are executed that may change the
        # visiblities. So it is critical to do the set_current_widget at the end.
        self.signal_set_current_widget.emit(self.stacked_widget.widget(index))

    def set_custom_widget(self, index: int, widget: QWidget) -> None:
        """Sets the custom widget for the specified step.

        Parameters:
        step (int): The step number to set the custom widget for.
        widget (QWidget): The custom widget to be used for the step.
        """
        current_idx = self.stacked_widget.currentIndex()

        # Remove the old widget if there is one
        old_widget = self.stacked_widget.widget(index)
        if old_widget is not None:
            self.signal_widget_unfocus.emit(old_widget)
            self.stacked_widget.removeWidget(old_widget)

        # Add the new custom widget for the step
        self.stacked_widget.insertWidget(index, widget)

        self.stacked_widget.setCurrentIndex(current_idx)

        # if the current active widget is changed, emit the signals,
        # as if it was switched to
        if index == current_idx:
            self.signal_set_current_widget.emit(widget)
            self.signal_widget_focus.emit(widget)


@dataclass
class VisibilityOption:
    widget: QWidget
    on_focus_set_visible: bool
    on_unfocus_set_visible: Optional[bool] = None


class TutorialWidget(QWidget):
    def __init__(
        self,
        container: StepProgressContainer,
        widget: QWidget,
        button_box: QDialogButtonBox,
        buttonbox_always_visible=False,
    ) -> None:
        super().__init__()
        self.container = container
        self.button_box = button_box
        self.visibility_options: List[VisibilityOption] = []
        self.buttonbox_always_visible = buttonbox_always_visible
        self.callback_on_set_current_widget: Optional[Callable] = None

        self._layout = QVBoxLayout(self)
        current_margins = self._layout.contentsMargins()
        self._layout.setContentsMargins(5, current_margins.top(), 5, 5)  # Left, Top, Right, Bottom margins

        self._layout.addWidget(widget)
        self._layout.addWidget(button_box)

        self.container.signal_set_current_widget.connect(self.on_set_current_widget)
        self.container.signal_widget_unfocus.connect(self.on_widget_unfocus)
        self.container.signal_widget_focus.connect(self.on_widget_focus)

    def set_widget(self, widget: QWidget) -> None:
        # Check if there is at least one widget in the layout
        if self._layout.count() > 0:
            # Take the first item (widget) from the layout
            item = self._layout.takeAt(0)
            if item is not None:
                # Remove the widget from the layout and delete it
                w = item.widget()
                if w is not None:  # Check if the item is a widget
                    w.setParent(None)  # type: ignore[call-overload]
                    w.deleteLater()  # Ensure the widget is deleted

        # Insert the new widget at position 0 in the layout
        self._layout.insertWidget(0, widget)

    def on_widget_unfocus(self, widget: QWidget) -> None:
        if self != widget:
            return
        logger.debug(f"on_widget_unfocus ")
        logger.debug(f"unset visibility of {self.visibility_options}")
        for visibility_option in self.visibility_options:
            if visibility_option.on_unfocus_set_visible is not None:
                visibility_option.widget.setVisible(visibility_option.on_unfocus_set_visible)
                logger.debug(
                    f"Setting {visibility_option.widget.__class__.__name__}.setVisible({visibility_option.on_unfocus_set_visible})"
                )

    def on_widget_focus(self, widget: QWidget) -> None:
        if self != widget:
            return
        logger.debug(f"on_widget_focus ")
        active_eq_visible = (
            self.container.step_bar.current_index == self.container.stacked_widget.currentIndex()
        )
        self.button_box.setVisible(active_eq_visible or self.buttonbox_always_visible)

        logger.debug(f"set visibility of {self.visibility_options}")
        for visibility_option in self.visibility_options:
            visibility_option.widget.setVisible(visibility_option.on_focus_set_visible)
            logger.debug(
                f"Setting {visibility_option.widget.__class__.__name__}.setVisible({visibility_option.on_focus_set_visible})"
            )

    def on_set_current_widget(self, widget: QWidget) -> None:
        if self != widget:
            return

        # the callbacks should only be in on_set_current_widget, but not in on_widget_focus (when i click on it)
        if self.callback_on_set_current_widget:
            logger.debug(f"on_set_current_widget: doing callback: {self.callback_on_set_current_widget}")
            self.callback_on_set_current_widget()

    def synchronize_visiblity(self, visibility_option: VisibilityOption) -> None:
        existing_widgets = [t.widget for t in self.visibility_options]
        if visibility_option.widget in existing_widgets:
            # handle the case that I set the visibility before
            idx = existing_widgets.index(visibility_option.widget)
            self.visibility_options[idx] = visibility_option
        else:
            self.visibility_options.append(visibility_option)

    def set_callback(self, callback: Callable) -> None:
        self.callback_on_set_current_widget = callback


class StepProgressContainerWithButtons(StepProgressContainer):
    def __init__(
        self,
        step_labels: List[str],
        signals_min: SignalsMin,
        current_index: int = 0,
        collapsible_current_active=False,
        clickable=True,
        use_checkmark_icon=True,
        parent=None,
        sub_indices: List[int] | None = None,
        use_resizing_stacked_widget=True,
        threading_parent: ThreadingManager | None = None,
    ) -> None:
        super().__init__(
            signals_min=signals_min,
            step_labels=step_labels,
            current_index=current_index,
            collapsible_current_active=collapsible_current_active,
            clickable=clickable,
            use_checkmark_icon=use_checkmark_icon,
            parent=parent,
            sub_indices=sub_indices,
            use_resizing_stacked_widget=use_resizing_stacked_widget,
            threading_parent=threading_parent,
        )

        for i in range(len(step_labels)):
            super().set_custom_widget(i, self.create_tutorial_widget())

    def set_custom_widget(self, index: int, widget: QWidget) -> None:
        tutorial_widget = self.stacked_widget.widget(index)
        if not isinstance(tutorial_widget, TutorialWidget):
            logger.error(f"{tutorial_widget} doesnt have the type TutorialWidget")
            return
        tutorial_widget.set_widget(widget)

        # if the current active widget is changed, emit the signals,
        # as if it was switched to
        if index == self.current_index():
            self.signal_set_current_widget.emit(tutorial_widget)
            self.signal_widget_focus.emit(tutorial_widget)

    def go_to_next_index(self) -> None:
        if self.current_index() + 1 < self.count():
            self.set_current_index(self.current_index() + 1)
        else:
            self.step_bar.set_mark_current_step_as_completed(True)

    def go_to_previous_index(self) -> None:
        self.step_bar.set_mark_current_step_as_completed(False)

        if self.current_index() - 1 >= 0:
            self.set_current_index(self.current_index() - 1)

    def create_tutorial_widget(self) -> TutorialWidget:
        widget = QWidget()
        QVBoxLayout(widget)

        buttonbox, buttons = create_button_box(
            self.go_to_next_index,
            self.go_to_previous_index,
            ok_text="Next step",
            cancel_text="Previous step",
        )
        return TutorialWidget(self, widget, buttonbox, buttonbox_always_visible=False)


if __name__ == "__main__":

    class DemoApp(QWidget):
        def __init__(self) -> None:
            super().__init__()

            # Create the StepProgressContainer with the desired steps
            self.step_progress_container = StepProgressContainerWithButtons(
                step_labels=["Create Account\n from hardware signers", "Login", "Payment", "Confirm"],
                current_index=1,
                sub_indices=[0, 2],
                signals_min=SignalsMin(),
                threading_parent=None,
            )
            self.step_progress_container.step_bar.set_enumeration_alphabet(["1.1", "1", "2.1", "2"])

            self.step_progress_container.step_bar.set_step_tooltips(
                [
                    "<font color='red'>This is an important step.</font><br><br><u>Payment information is required.</u><br><br><b>Confirm your submission.</b>",
                    "<i>Remember to check your details.</i>",
                    "<u>Payment information is required.</u>",
                    "<b>Confirm your submission.</b>",
                ]
            )

            def factory(i) -> Callable:
                def f(i=i) -> None:
                    print(f"callback action for {i}")

                return f

            for i in range(self.step_progress_container.count()):
                widget = self.step_progress_container.stacked_widget.widget(i)
                if not isinstance(widget, TutorialWidget):
                    continue
                widget.set_callback(factory(i))
                self.step_progress_container.set_custom_widget(i, QTextEdit(f"{i}"))

            self.init_ui()

        def init_ui(self) -> None:
            self._layout = QVBoxLayout(self)

            # Buttons to navigate through steps
            next_button = QPushButton("Next Step")
            next_button.clicked.connect(self.next_index)

            prev_button = QPushButton("Previous Step")
            prev_button.clicked.connect(self.prev_index)

            self._layout.addWidget(
                self.step_progress_container
            )  # Add the step progress container instead of step_bar
            self._layout.addWidget(prev_button)
            self._layout.addWidget(next_button)
            self._layout.setContentsMargins(0, 0, 0, 0)

            self.toggle_completion_button = QPushButton("Toggle Step Completion")
            self.toggle_completion_button.clicked.connect(self.toggle_step_completion)
            self._layout.addWidget(self.toggle_completion_button)

        def toggle_step_completion(self) -> None:
            current_value = self.step_progress_container.step_bar.mark_current_index_as_completed
            self.step_progress_container.step_bar.set_mark_current_step_as_completed(not current_value)

        def next_index(self) -> None:
            if (
                self.step_progress_container.step_bar.current_index
                < self.step_progress_container.step_bar.number_of_steps
            ):
                self.step_progress_container.set_current_index(
                    self.step_progress_container.step_bar.current_index + 1
                )

        def prev_index(self) -> None:
            if self.step_progress_container.step_bar.current_index > 0:
                self.step_progress_container.set_current_index(
                    self.step_progress_container.step_bar.current_index - 1
                )

    app = QApplication(sys.argv)
    window = DemoApp()
    window.show()
    sys.exit(app.exec())



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Original Version from:
#
# Electrum - lightweight Bitcoin client
# Copyright (C) 2023 The Electrum Developers
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging

from bitcoin_safe.gui.qt.dialog_import import file_to_str
from bitcoin_safe.gui.qt.html_delegate import HTMLDelegate
from bitcoin_safe.gui.qt.wrappers import Menu
from bitcoin_safe.signals import Signals
from bitcoin_safe.util import str_to_qbytearray
from bitcoin_safe.wallet import TxStatus

from ...config import UserConfig
from ...i18n import translate

logger = logging.getLogger(__name__)

import csv
import enum
import io
import json
import os
import os.path
import tempfile
from decimal import Decimal
from typing import Any, Callable, Dict, Iterable, List, Optional, Sequence, Type, Union

from bitcoin_qr_tools.data import Data
from PyQt6 import QtCore
from PyQt6.QtCore import (
    QAbstractItemModel,
    QEvent,
    QItemSelection,
    QItemSelectionModel,
    QMimeData,
    QModelIndex,
    QPersistentModelIndex,
    QPoint,
    QSize,
    QSortFilterProxyModel,
    Qt,
    QUrl,
    pyqtSignal,
)
from PyQt6.QtGui import (
    QAction,
    QCursor,
    QDrag,
    QDragEnterEvent,
    QDragMoveEvent,
    QDropEvent,
    QFont,
    QHelpEvent,
    QKeyEvent,
    QMouseEvent,
    QPainter,
    QPixmap,
    QRegion,
    QShowEvent,
    QStandardItem,
    QStandardItemModel,
)
from PyQt6.QtWidgets import (
    QAbstractItemView,
    QFileDialog,
    QHBoxLayout,
    QHeaderView,
    QLabel,
    QLineEdit,
    QMenu,
    QStyledItemDelegate,
    QStyleOptionViewItem,
    QToolButton,
    QTreeView,
    QVBoxLayout,
    QWidget,
)

from .util import do_copy, read_QIcon


def needs_frequent_flag(status: TxStatus | None) -> bool:
    if not status:
        return True

    if status.confirmations() < 6:
        return True
    return False


class MyItemDataRole(enum.IntEnum):
    ROLE_CLIPBOARD_DATA = Qt.ItemDataRole.UserRole + 100
    ROLE_CUSTOM_PAINT = Qt.ItemDataRole.UserRole + 101
    ROLE_EDIT_KEY = Qt.ItemDataRole.UserRole + 102
    ROLE_FILTER_DATA = Qt.ItemDataRole.UserRole + 103
    ROLE_SORT_ORDER = Qt.ItemDataRole.UserRole + 1000
    ROLE_KEY = Qt.ItemDataRole.UserRole + 1001
    ROLE_FREQUENT_UPDATEFLAG = Qt.ItemDataRole.UserRole + 1002


class MyMenu(Menu):
    def __init__(self, config: UserConfig) -> None:
        QMenu.__init__(self)
        self.setToolTipsVisible(True)
        self.config = config

    def addToggle(self, text: str, callback: Callable, *, tooltip="") -> QAction:
        m = self.add_action(text, callback)
        m.setCheckable(True)
        m.setToolTip(tooltip)
        return m


class MyStandardItemModel(QStandardItemModel):
    def __init__(
        self,
        parent,
        drag_key: str = "item",
        drag_keys_to_file_paths=None,
    ) -> None:
        super().__init__(parent)
        self.mytreeview: MyTreeView = parent
        self.drag_key = drag_key
        self.drag_keys_to_file_paths = self.csv_drag_keys_to_file_paths
        if drag_keys_to_file_paths:
            self.drag_keys_to_file_paths = drag_keys_to_file_paths

    def csv_drag_keys_to_file_paths(
        self, drag_keys: Iterable[str], save_directory: Optional[str] = None
    ) -> List[str]:
        """Writes the selected rows in a csv file (the directory is )"""
        file_path = os.path.join(save_directory, f"export.csv") if save_directory else None
        return [self.mytreeview.csv_drag_keys_to_file_path(drag_keys=drag_keys, file_path=file_path)]

    def flags(self, index: QtCore.QModelIndex) -> QtCore.Qt.ItemFlag:
        if index.column() == self.mytreeview.key_column:  # only enable dragging for column 1
            return super().flags(index) | Qt.ItemFlag.ItemIsDragEnabled
        else:
            return super().flags(index)

    def mimeData(self, indexes: Iterable[QtCore.QModelIndex]) -> QMimeData:
        mime_data = QMimeData()
        keys = set()
        for index in indexes:
            if index.isValid():
                item = self.item(index.row(), self.mytreeview.key_column)
                if not item:
                    continue
                key = item.data(role=MyItemDataRole.ROLE_KEY)
                keys.add(key)

        # set the key data for internal drags
        d = {
            "type": f"drag_{self.drag_key}",
            self.drag_key: list(keys),
        }

        mime_data.setData("application/json", str_to_qbytearray(json.dumps(d)))

        # set the key data for files

        file_urls = []
        for file_path in self.drag_keys_to_file_paths(keys):
            # Add the file URL to the list
            file_urls.append(QUrl.fromLocalFile(file_path))

        # Set the URLs of the files in the mime data
        mime_data.setUrls(file_urls)

        return mime_data


class MySortModel(QSortFilterProxyModel):
    def __init__(self, parent, source_model: MyStandardItemModel, sort_role: int) -> None:
        super().__init__(parent)
        self._sort_role = sort_role
        self._source_model = source_model
        self.setSourceModel(source_model)
        self.setSortRole(sort_role)

    def setSourceModel(self, sourceModel: MyStandardItemModel) -> None:  # type: ignore[override]
        self._source_model = sourceModel
        super().setSourceModel(sourceModel)

    def sourceModel(self) -> MyStandardItemModel:
        if self._source_model:
            return self._source_model
        return MyStandardItemModel(parent=self)

    def lessThan(self, source_left: QModelIndex, source_right: QModelIndex) -> bool:
        item1 = self.sourceModel().itemFromIndex(source_left)
        item2 = self.sourceModel().itemFromIndex(source_right)

        if not item1 or not item2:
            return bool(item1) < bool(item2)

        data1 = item1.data(self._sort_role)
        data2 = item2.data(self._sort_role)
        if data1 is not None and data2 is not None:
            return data1 < data2
        v1 = item1.text()
        v2 = item2.text()
        try:
            return Decimal(v1) < Decimal(v2)
        except:
            return v1 < v2


class ElectrumItemDelegate(QStyledItemDelegate):
    def __init__(self, tv: "MyTreeView") -> None:
        super().__init__(tv)
        self.icon_shift_right = 30
        self.tv = tv
        self.opened: Optional[QPersistentModelIndex] = None

        def on_closeEditor(editor: QLineEdit, hint) -> None:
            self.opened = None
            self.tv.is_editor_open = False
            if self.tv._pending_update:
                self.tv.update_content()

        def on_commitData(editor: QLineEdit) -> None:
            new_text = editor.text()
            if not self.opened:
                return
            idx = QModelIndex(self.opened)
            row, col = idx.row(), idx.column()
            edit_key = self.tv.get_edit_key_from_coordinate(row, col)
            assert edit_key is not None, (idx.row(), idx.column())
            self.tv.on_edited(idx, edit_key=edit_key, text=new_text)

        self.closeEditor.connect(on_closeEditor)
        self.commitData.connect(on_commitData)

    def initStyleOption(self, option: QStyleOptionViewItem | None, index: QModelIndex) -> None:
        super().initStyleOption(option, index)
        if not option:
            return
        option.displayAlignment = self.tv.column_alignments.get(index.column(), Qt.AlignmentFlag.AlignLeft)

    def createEditor(
        self, parent: Optional[QWidget], option: QStyleOptionViewItem, index: QtCore.QModelIndex
    ) -> Optional[QWidget]:
        self.opened = QPersistentModelIndex(index)
        self.tv.is_editor_open = True
        return super().createEditor(parent, option, index)

    def paint(self, painter: QPainter | None, option: QStyleOptionViewItem, idx: QModelIndex) -> None:
        custom_data = idx.data(MyItemDataRole.ROLE_CUSTOM_PAINT)
        if isinstance(custom_data, HTMLDelegate):
            custom_data.paint(painter, option, idx)
        else:
            super().paint(painter, option, idx)

    def helpEvent(
        self,
        evt: QHelpEvent | None,
        view: QAbstractItemView | None,
        option: QStyleOptionViewItem,
        idx: QModelIndex,
    ) -> bool:
        custom_data = idx.data(MyItemDataRole.ROLE_CUSTOM_PAINT)
        if custom_data is None:
            return super().helpEvent(evt, view, option, idx)
        else:
            if evt and evt.type() == QEvent.Type.ToolTip and isinstance(custom_data, HTMLDelegate):
                if custom_data.show_tooltip(evt):
                    return True
        return super().helpEvent(evt, view, option, idx)

    def sizeHint(self, option: QStyleOptionViewItem, idx: QModelIndex) -> QSize:
        custom_data = idx.data(MyItemDataRole.ROLE_CUSTOM_PAINT)
        if custom_data is None:
            return super().sizeHint(option, idx)
        else:
            # default_size = super().sizeHint(option, idx)
            return custom_data.sizeHint(option, idx)


class MyTreeView(QTreeView):
    signal_selection_changed = pyqtSignal()
    signal_update = pyqtSignal()

    filter_columns: Iterable[int]
    column_alignments: Dict[int, Qt.AlignmentFlag] = {}
    hidden_columns: List[int] = []

    key_column = 0

    class BaseColumnsEnum(enum.IntEnum):
        @staticmethod
        def _generate_next_value_(name: str, start: int, count: int, last_values) -> int:
            # this is overridden to get a 0-based counter
            return count

    Columns: Type[BaseColumnsEnum]

    def __init__(
        self,
        *,
        config: UserConfig,
        signals: Signals,
        parent: Optional[QWidget] = None,
        stretch_column: Optional[int] = None,
        column_widths: Optional[Dict[BaseColumnsEnum, int]] = None,
        editable_columns: Optional[Sequence[int]] = None,
        sort_column: int | None = None,
        sort_order: Qt.SortOrder = Qt.SortOrder.AscendingOrder,
    ) -> None:
        parent = parent
        super().__init__(parent)
        self.signals = signals
        self._source_model = MyStandardItemModel(parent)
        self.config = config
        self.stretch_column = stretch_column
        self.column_widths = column_widths if column_widths else {}
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.create_menu)
        self.setUniformRowHeights(True)

        # Control which columns are editable
        if editable_columns is None:
            editable_columns = []
        self.editable_columns = set(editable_columns)
        self.setItemDelegate(ElectrumItemDelegate(self))
        self.current_filter = ""
        self.is_editor_open = False
        self._currently_updating = False
        self._scroll_position = 0

        self.setRootIsDecorated(False)  # remove left margin

        # When figuring out the size of columns, Qt by default looks at
        # the first 1000 rows (at least if resize mode is QHeaderView.ResizeToContents).
        # This would be REALLY SLOW, and it's not perfect anyway.
        # So to speed the UI up considerably, set it to
        # only look at as many rows as currently visible.
        if isinstance(header := self.header(), QHeaderView):
            header.setResizeContentsPrecision(0)
        self._pending_update = False
        self._forced_update = False

        self._default_bg_brush = QStandardItem().background()
        self.proxy = QSortFilterProxyModel()

        # Here's where we set the font globally for the view
        font = QFont("Arial", 10)
        self.setFont(font)

        self.setAcceptDrops(True)
        if viewport := self.viewport():
            viewport.setAcceptDrops(True)
        self.setDropIndicatorShown(True)
        self.setDragDropMode(QAbstractItemView.DragDropMode.DragDrop)
        self.setDefaultDropAction(Qt.DropAction.CopyAction)
        self.setDragEnabled(True)  # this must be after the other drag toggles

        self._current_column = sort_column if sort_column is not None else self.key_column
        self._current_order = sort_order
        self.sortByColumn(self._current_column, self._current_order)

    def setItemDelegate(self, delegate: ElectrumItemDelegate) -> None:  # type: ignore[override]
        self._item_delegate = delegate
        super().setItemDelegate(delegate)

    def updateUi(self) -> None:
        pass

    def startDrag(self, action: Qt.DropAction) -> None:
        indexes = self.selectedIndexes()
        if indexes:
            drag = QDrag(self)
            mime_data = self.model().mimeData(indexes)
            drag.setMimeData(mime_data)

            total_height = sum(self.visualRect(index).height() for index in indexes)
            max_width = max(self.visualRect(index).width() for index in indexes)

            pixmap = QPixmap(max_width, total_height)
            pixmap.fill(Qt.GlobalColor.transparent)

            painter = QPainter(pixmap)
            current_height = 0
            for index in indexes:
                if index.column() != self.key_column:
                    continue
                rect = self.visualRect(index)
                temp_pixmap = QPixmap(rect.size())
                if viewport := self.viewport():
                    viewport.render(temp_pixmap, QPoint(), QRegion(rect))
                painter.drawPixmap(0, int(current_height), temp_pixmap)
                current_height += rect.height()
            painter.end()

            # self.mapFromGlobal(QCursor.pos())
            # self.visualRect(indexes[0]).bottomLeft()
            hotspot_pos = QPoint(0, 0)  # cursor_pos - visual_rect
            # the y offset is always off, so just set it completely to 0
            hotspot_pos.setY(0)
            drag.setPixmap(pixmap)
            drag.setHotSpot(hotspot_pos)

            drag.exec(action)

    def create_menu(self, position: QPoint) -> Menu:
        menu = Menu()
        selected: List[QModelIndex] = self.selected_in_column(self.key_column)
        if not selected:
            current_row = self.current_row_in_column(self.key_column)
            if current_row:
                selected = [current_row]

        if not selected:
            return menu

        multi_select = len(selected) > 1
        if not multi_select:
            idx = self.indexAt(position)
            if not idx.isValid():
                return menu

            self.add_copy_menu(menu, idx, include_columns_even_if_hidden=[self.key_column])

        menu.add_action(
            self.tr("Copy as csv"),
            lambda: self.copyRowsToClipboardAsCSV([r.row() for r in selected]),
            icon=read_QIcon("csv-file.svg"),
        )

        # run_hook('receive_menu', menu, addrs, self.wallet)
        if viewport := self.viewport():
            menu.exec(viewport.mapToGlobal(position))

        return menu

    def add_copy_menu(self, menu: Menu, idx: QModelIndex, include_columns_even_if_hidden=None) -> Menu:
        copy_menu = menu.add_menu(self.tr("Copy"))
        copy_menu.setIcon(read_QIcon("copy.png"))

        def factory(text, title):
            def f(text=text, title=title):
                self.place_text_on_clipboard(text=text, title=title)

            return f

        for column in self.Columns:
            if self.isColumnHidden(column) and (
                include_columns_even_if_hidden is None or column not in include_columns_even_if_hidden
            ):
                continue
            item = self.sourceModel().horizontalHeaderItem(column)
            if not item:
                continue
            column_title = item.text()
            if not column_title:
                continue
            item_col = self.item_from_index(idx.sibling(idx.row(), column))
            if not item_col:
                continue
            clipboard_data = item_col.data(MyItemDataRole.ROLE_CLIPBOARD_DATA)
            if clipboard_data is None:
                clipboard_data = item_col.text().strip()
            if not clipboard_data:
                continue

            copy_menu.add_action(column_title, factory(text=clipboard_data, title=column_title))
        return copy_menu

    def set_editability(self, items: List[QStandardItem]) -> None:
        for idx, i in enumerate(items):
            i.setEditable(idx in self.editable_columns)

    def selected_in_column(self, column: int) -> List[QModelIndex]:
        selection_model = self.selectionModel()
        if not selection_model:
            return []
        items = selection_model.selectedIndexes()
        return list(x for x in items if x.column() == column)

    def current_row_in_column(self, column: int) -> Optional[QModelIndex]:
        selection_model = self.selectionModel()
        if not selection_model:
            return None

        idx = selection_model.currentIndex()
        if idx.isValid():
            # Retrieve data for a specific role from the current index
            # Replace 'YourSpecificRole' with the role you are interested in
            # For example, QtCore.Qt.DisplayRole for the display text
            return idx.sibling(idx.row(), column)
        return None

    def get_role_data_for_current_item(self, *, col, role) -> Any:
        selection_model = self.selectionModel()
        if not selection_model:
            return None

        idx = selection_model.currentIndex()
        idx = idx.sibling(idx.row(), col)
        item = self.item_from_index(idx)
        if item:
            return item.data(role)

    def model(self) -> QAbstractItemModel:
        if self.proxy:
            return self.proxy
        return self._source_model

    def sourceModel(self) -> MyStandardItemModel:
        return self._source_model

    def itemDelegate(self) -> ElectrumItemDelegate:
        return self._item_delegate

    def item_from_index(self, idx: QModelIndex) -> Optional[QStandardItem]:
        model = self.model()
        if isinstance(model, QSortFilterProxyModel):
            return self.sourceModel().itemFromIndex(model.mapToSource(idx))
        else:
            return self.sourceModel().itemFromIndex(idx)

    def set_current_idx(self, set_current: QPersistentModelIndex) -> None:
        if not set_current or not set_current.isValid():
            return
        selection_model = self.selectionModel()
        if not selection_model:
            return
        selection_model.select(QModelIndex(set_current), QItemSelectionModel.SelectionFlag.SelectCurrent)

    def select_row(self, content, column, role: MyItemDataRole = MyItemDataRole.ROLE_KEY) -> None:
        return self.select_rows([content], column, role)

    def select_rows(
        self,
        content_list,
        column,
        role: MyItemDataRole = MyItemDataRole.ROLE_KEY,
        clear_previous_selection=True,
    ) -> None:
        last_selected_index = None
        model = self.model()
        selection_model = self.selectionModel()
        if not selection_model:
            return

        self._currently_updating = True

        if clear_previous_selection:
            selection_model.clear()  # Clear previous selection
        for row in range(model.rowCount()):
            index = model.index(row, column)
            this_content = model.data(index, role)
            if this_content in content_list:
                # Select the item
                selection_model.select(
                    index, QItemSelectionModel.SelectionFlag.Select | QItemSelectionModel.SelectionFlag.Rows
                )
                last_selected_index = index
        if last_selected_index:
            # Scroll to the last selected index
            self.scrollTo(last_selected_index)

        if viewport := self.viewport():
            viewport.update()
        self._currently_updating = False

        self.signal_selection_changed.emit()

    def update_headers(
        self,
        headers: Union[Dict[Any, str], Iterable[str]],
    ) -> None:
        if not isinstance(header := self.header(), QHeaderView):
            return
        # Get the current sorting column and order
        current_column = header.sortIndicatorSection()
        current_order = header.sortIndicatorOrder()

        # headers is either a list of column names, or a dict: (col_idx->col_name)
        if not isinstance(headers, dict):  # convert to dict
            headers = dict(enumerate(headers))
        col_names = [headers[col_idx] for col_idx in sorted(headers.keys())]
        self.sourceModel().setHorizontalHeaderLabels(col_names)
        header.setSortIndicator(current_column, current_order)
        self.sortByColumn(current_column, current_order)
        header.setStretchLastSection(False)
        for col_idx in headers:
            sm = (
                QHeaderView.ResizeMode.Stretch
                if col_idx == self.stretch_column or col_idx in self.column_widths.keys()
                else QHeaderView.ResizeMode.ResizeToContents
            )
            header.setSectionResizeMode(int(col_idx), sm)

        for col_idx, width in self.column_widths.items():
            header.setSectionResizeMode(col_idx, QHeaderView.ResizeMode.Interactive)
            header.resizeSection(col_idx, width)

    def selectionChanged(self, selected: QItemSelection, deselected: QItemSelection) -> None:
        super().selectionChanged(selected, deselected)
        if self._currently_updating:
            return
        self.signal_selection_changed.emit()

    def keyPressEvent(self, event: QKeyEvent | None) -> None:
        if self.itemDelegate().opened:
            return  # type: ignore[unreachable]

        selection_model = self.selectionModel()
        if not event or not selection_model:
            super().keyPressEvent(event)
            return

        if event.key() in [Qt.Key.Key_Return, Qt.Key.Key_Enter]:
            self.on_activated(selection_model.currentIndex())
            return
        if event.key() in [Qt.Key.Key_F2]:
            if not self.editable_columns:
                return
            idx = selection_model.currentIndex()
            idx = idx.sibling(idx.row(), list(self.editable_columns)[0])
            self.edit(
                QModelIndex(QPersistentModelIndex(idx)), QAbstractItemView.EditTrigger.AllEditTriggers, event
            )
            return

        if (event.modifiers() & Qt.KeyboardModifier.ControlModifier) and (event.key() == Qt.Key.Key_C):
            selection = selection_model.selection().indexes()
            if selection:
                self.copyKeyRoleToClipboard(set([index.row() for index in selection]))
        else:
            super().keyPressEvent(event)

    def copyKeyRoleToClipboard(self, row_numbers) -> None:
        def get_data(row, col) -> Any:
            model = self.model()
            index = model.index(row, self.key_column)

            if hasattr(model, "data"):
                key = model.data(index, MyItemDataRole.ROLE_KEY)
                return key
            else:
                item = self.item_from_index(index)
                if item:
                    key = item.data(MyItemDataRole.ROLE_KEY)
                    return key

        row_numbers = sorted(row_numbers)

        stream = io.StringIO()
        for row in row_numbers:
            stream.write(
                str(get_data(row, MyItemDataRole.ROLE_KEY)) + "\n"
            )  # append newline character after each row
        do_copy(stream.getvalue(), title=f"{len(row_numbers)} rows have been copied as text")

    def get_rows_as_list(self, row_numbers) -> Any:
        def get_data(row, col) -> Any:
            model = self.model()  # assuming this is a QAbstractItemModel or subclass
            index = model.index(row, col)

            if hasattr(model, "data"):
                return model.data(index, MyItemDataRole.ROLE_CLIPBOARD_DATA)
            else:
                item = self.item_from_index(index)
                if item:
                    return item.data(MyItemDataRole.ROLE_CLIPBOARD_DATA)

        row_numbers = sorted(row_numbers)

        table = []
        headers = [
            self.model().headerData(i, QtCore.Qt.Orientation.Horizontal)
            for i in range(self.model().columnCount())
        ]  # retrieve headers
        table.append(headers)  # write headers to table

        for row in row_numbers:
            row_data = []
            for column in self.Columns:
                row_data.append(get_data(row, column))
            table.append(row_data)

        return table

    def copyRowsToClipboardAsCSV(self, row_numbers) -> None:
        table = self.get_rows_as_list(row_numbers)

        stream = io.StringIO()
        writer = csv.writer(stream)
        writer.writerows(table)
        do_copy(stream.getvalue(), title=f"{len(row_numbers)} rows have ben copied as csv")

    def mouseDoubleClickEvent(self, event: QMouseEvent | None) -> None:
        if not event:
            return
        idx: QModelIndex = self.indexAt(event.pos())
        if self.proxy:
            idx = self.proxy.mapToSource(idx)
        if not idx.isValid():
            # can happen e.g. before list is populated for the first time
            return

        if idx.column() in self.editable_columns:
            super().mouseDoubleClickEvent(event)
        else:
            self.on_double_click(idx)

    def on_double_click(self, idx: QModelIndex) -> None:
        pass

    def on_activated(self, idx: QModelIndex) -> None:
        # on 'enter' we show the menu
        pt = self.visualRect(idx).bottomLeft()
        pt.setX(50)
        self.customContextMenuRequested.emit(pt)

    def on_edited(self, idx: QModelIndex, edit_key, *, text: str) -> None:
        raise NotImplementedError()

    def get_text_from_coordinate(self, row: int, col: int) -> str:
        idx = self.model().index(row, col)
        item = self.item_from_index(idx)
        if not item:
            return ""
        return item.text()

    def get_role_data_from_coordinate(self, row: int, col: int, *, role) -> Any:
        idx = self.model().index(row, col)
        item = self.item_from_index(idx)
        if not item:
            return None
        role_data = item.data(role)
        return role_data

    def get_edit_key_from_coordinate(self, row: int, col: int) -> Any:
        # overriding this might allow avoiding storing duplicate data
        return self.get_role_data_from_coordinate(row, col, role=MyItemDataRole.ROLE_EDIT_KEY)

    def get_filter_data_from_coordinate(self, row: int, col: int) -> str:
        filter_data = self.get_role_data_from_coordinate(row, col, role=MyItemDataRole.ROLE_FILTER_DATA)
        if filter_data:
            return filter_data
        txt: str = self.get_text_from_coordinate(row, col)
        txt = txt.lower()
        return txt

    def hide_row(self, row_num: int) -> bool:
        """row_num is for self.model(). So if there is a proxy, it is the row
        number in that!

        It returns:  is_now_hidden
        """
        is_now_hidden = False
        if not self.current_filter:
            # no filters at all, neither date nor search
            is_now_hidden = False
            self.setRowHidden(row_num, QModelIndex(), is_now_hidden)
            return is_now_hidden
        for column in self.filter_columns:
            filter_data = self.get_filter_data_from_coordinate(row_num, column)
            if self.current_filter in filter_data:
                # the filter matched, but the date filter might apply
                self.setRowHidden(row_num, QModelIndex(), False)
                break
        else:
            # we did not find the filter in any columns, hide the item
            is_now_hidden = True
            self.setRowHidden(row_num, QModelIndex(), True)
        return is_now_hidden

    def filter(self, p=None) -> List[bool]:
        "Returns a [row0_is_now_hidden, row1_is_now_hidden, ...]"
        if p is not None:
            p = p.lower()
            self.current_filter = p
        return self.hide_rows()

    def hide_rows(self) -> List[bool]:
        "Returns a [row0_is_now_hidden, row1_is_now_hidden, ...]"
        return [self.hide_row(row) for row in range(self.model().rowCount())]

    def as_csv_string(self, row_numbers: Optional[List[int]] = None, export_all=False) -> str:
        table = self.get_rows_as_list(
            row_numbers=list(range(self.model().rowCount())) if export_all else row_numbers
        )

        stream = io.StringIO()
        writer = csv.writer(stream)
        writer.writerows(table)

        return stream.getvalue()

    def export_as_csv(self, file_path=None) -> None:
        if not file_path:
            file_path, _ = QFileDialog.getSaveFileName(
                self, self.tr("Export csv"), "", self.tr("All Files (*);;Text Files (*.csv)")
            )
            if not file_path:
                logger.info("No file selected")
                return

        self.csv_drag_keys_to_file_path(file_path=file_path, export_all=True)

    def csv_drag_keys_to_file_path(
        self, drag_keys: Optional[Iterable[str]] = None, file_path: str | None = None, export_all=False
    ) -> str:
        row_numbers: List[int] = []
        if drag_keys and not export_all:
            for row_number in range(0, self._source_model.rowCount()):
                item = self._source_model.item(row_number, self.key_column)
                if item and item.data(MyItemDataRole.ROLE_KEY) in drag_keys:
                    row_numbers.append(row_number)

        # Fetch the serialized data using the drag_keys
        csv_string = self.as_csv_string(row_numbers=row_numbers, export_all=export_all)

        if file_path:
            file_descriptor = os.open(file_path, os.O_CREAT | os.O_WRONLY)
        else:
            # Create a temporary file
            file_descriptor, file_path = tempfile.mkstemp(
                suffix=f".csv",
                prefix=f"{self._source_model.drag_key} ",
            )

        with os.fdopen(file_descriptor, "w") as file:
            file.write(csv_string)

        logger.debug(f"CSV Table saved to {file_path}")
        return file_path

    def place_text_on_clipboard(self, text: str, *, title: str | None = None) -> None:
        do_copy(text, title=title)

    def showEvent(self, e: QShowEvent | None) -> None:
        super().showEvent(e)
        if e and e.isAccepted() and self._pending_update:
            self._forced_update = True
            self.update_content()
            self._forced_update = False

    def maybe_defer_update(self) -> bool:
        """Returns whether we should defer an update/refresh."""
        defer = not self._forced_update and (not self.isVisible() or self.is_editor_open)
        # side-effect: if we decide to defer update, the state will become stale:
        self._pending_update = defer
        return defer

    def find_row_by_key(self, key: str) -> Optional[int]:
        for row in range(0, self._source_model.rowCount()):
            item = self._source_model.item(row, self.key_column)
            if item and item.data(MyItemDataRole.ROLE_KEY) == key:
                return row
        return None

    def refresh_all(self) -> None:
        if self.maybe_defer_update():
            return
        for row in range(0, self._source_model.rowCount()):
            item = self._source_model.item(row, self.key_column)
            if not item:
                continue
            key = item.data(MyItemDataRole.ROLE_KEY)
            self.refresh_row(key, row)

    def refresh_row(self, key: str, row: int) -> None:
        pass

    def refresh_item(self, key: str) -> None:
        row = self.find_row_by_key(key)
        if row is not None:
            self.refresh_row(key, row)

    def delete_item(self, key: str) -> None:
        row = self.find_row_by_key(key)
        if row is not None:
            self._source_model.takeRow(row)

    @staticmethod
    def _recognized_files(mime_data: QMimeData) -> List[str]:
        result: List[str] = []
        if mime_data.hasUrls():
            # Iterate through the list of dropped file URLs
            for url in mime_data.urls():
                # Convert URL to local file path
                file_path = url.toLocalFile()
                if file_path.endswith(".wallet") or file_path.endswith(".tx") or file_path.endswith(".psbt"):
                    result.append(file_path)
        return result

    def dragEnterEvent(self, event: QDragEnterEvent | None) -> None:
        if not event:
            return
        if (mime_data := event.mimeData()) and self._recognized_files(mime_data):
            event.accept()
            return

        if not event.isAccepted():
            event.ignore()

    def dragMoveEvent(self, event: QDragMoveEvent | None) -> None:
        if not event:
            return
        if (mime_data := event.mimeData()) and self._recognized_files(mime_data):
            event.accept()
            return

        if not event.isAccepted():
            event.ignore()

    def dropEvent(self, event: QDropEvent | None) -> None:
        if not event:
            return
        mime_data = event.mimeData()
        if mime_data:
            file_paths = self._recognized_files(mime_data)
            for file_path in file_paths:
                if file_path.endswith(".wallet"):
                    logger.debug(file_path)
                    event.accept()
                    self.signals.open_wallet.emit(file_path)

                if file_path.endswith(".tx") or file_path.endswith(".psbt"):
                    logger.debug(file_path)
                    event.accept()

                    data = Data.from_str(file_to_str(file_path), self.config.network)
                    self.signals.open_tx_like.emit(data.data)

        if not event.isAccepted():
            event.ignore()

    def _save_selection(self):
        self.selected_ids = []
        selection_model = self.selectionModel()
        if not selection_model:
            return

        # Save the current scroll position
        scrollbar = self.verticalScrollBar()
        if scrollbar:
            self._scroll_position = scrollbar.value()

        selected_indexes = selection_model.selectedRows(self.key_column)
        for index in selected_indexes:
            if index.isValid():
                # Map the index to the source model if using a proxy
                source_index = self.proxy.mapToSource(index)
                id = source_index.data(MyItemDataRole.ROLE_KEY)
                if id is not None:
                    self.selected_ids.append(id)

    def _restore_selection(self):
        selection_model = self.selectionModel()
        if not selection_model:
            return

        selection_model.clearSelection()
        scrollbar = self.verticalScrollBar()
        if scrollbar:
            scrollbar.setValue(self._scroll_position)  # Restore the scroll position

        # Iterate through all rows in the model to find items with matching IDs
        for row in range(self._source_model.rowCount()):
            index = self._source_model.index(row, self.key_column)
            id = index.data(MyItemDataRole.ROLE_KEY)
            if id in self.selected_ids:
                # Map the source index to the proxy model
                proxy_index = self.proxy.mapFromSource(index)
                # Select the item
                selection_model.select(
                    proxy_index,
                    QItemSelectionModel.SelectionFlag.Select | QItemSelectionModel.SelectionFlag.Rows,
                )

    def _before_update_content(self):
        self._currently_updating = True
        self._save_selection()
        if not isinstance(header := self.header(), QHeaderView):
            return
        self._current_column = header.sortIndicatorSection()
        self._current_order = header.sortIndicatorOrder()
        self.proxy.setDynamicSortFilter(False)  # temp. disable re-sorting after every change

    def _after_update_content(self):
        self.sortByColumn(self._current_column, self._current_order)
        self.proxy.setDynamicSortFilter(True)

        # show/hide self.Columns
        self.filter()

        for hidden_column in self.hidden_columns:
            self.hideColumn(hidden_column)

        self._restore_selection()
        # this MUST be after the selection,
        # such that on_selection_change is not triggered
        self._currently_updating = False

    def update_content(self) -> None:
        super().update()
        logger.debug(f"{self.__class__.__name__} done updating")
        # sort again just as before
        self.signal_update.emit()

    @staticmethod
    def get_json_mime_data(mime_data: QMimeData) -> Optional[Dict]:
        if mime_data.hasFormat("application/json"):
            data_bytes = mime_data.data("application/json")
            try:
                json_string = data_bytes.data().decode()
                logger.debug(f"dragEnterEvent: {json_string}")
                d = json.loads(json_string)
                return d
            except:
                return None

        return None


class SearchableTab(QWidget):

    def __init__(self, parent=None, **kwargs) -> None:
        super().__init__(parent=parent)

        self.searchable_list: MyTreeView


class TreeViewWithToolbar(SearchableTab):
    def __init__(
        self, searchable_list: MyTreeView, config: UserConfig, parent: QWidget | None = None
    ) -> None:
        super().__init__(parent=parent)
        self.config = config
        self.toolbar_is_visible = False
        self.searchable_list = searchable_list

        # signals
        self.searchable_list.signal_update.connect(self.update)
        # in searchable_list signal_update will be sent after the update. and since this
        # is relevant for the balance to show, i need to update also the balance label
        # which is done in updateUi
        self.searchable_list.signal_update.connect(self.updateUi)

    def create_layout(self) -> None:
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)  # Left, Top, Right, Bottom margins

        self.create_toolbar_with_menu("")
        layout.addLayout(self.toolbar)
        layout.addWidget(self.searchable_list)

    def create_toolbar_with_menu(self, title):
        self.menu = MyMenu(self.config)
        self.action_export_as_csv = self.menu.add_action(
            "", self.searchable_list.export_as_csv, icon=read_QIcon("csv-file.svg")
        )

        toolbar_button = QToolButton()

        def create_menu():
            self.menu.exec(QCursor.pos())

        toolbar_button.clicked.connect(create_menu)
        toolbar_button.setIcon(read_QIcon("preferences.svg"))
        toolbar_button.setMenu(self.menu)
        toolbar_button.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)
        toolbar_button.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.toolbar = QHBoxLayout()

        self.balance_label = QLabel()

        self.search_edit = QLineEdit()
        self.search_edit.setClearButtonEnabled(True)
        self.search_edit.textChanged.connect(self.searchable_list.filter)

        self.toolbar.addWidget(self.balance_label)
        self.toolbar.addStretch()
        self.toolbar.addWidget(self.search_edit)
        self.toolbar.addWidget(toolbar_button)
        return self.toolbar, self.menu, self.balance_label, self.search_edit, self.action_export_as_csv

    def show_toolbar(self, is_visible: bool, config=None) -> None:
        if is_visible == self.toolbar_is_visible:
            return
        self.toolbar_is_visible = is_visible
        if not is_visible:
            self.on_hide_toolbar()

    def on_hide_toolbar(self) -> None:
        pass

    def toggle_toolbar(self, config=None) -> None:
        self.show_toolbar(not self.toolbar_is_visible, config)

    def updateUi(self) -> None:
        self.search_edit.setPlaceholderText(translate("mytreeview", "Type to filter"))
        self.action_export_as_csv.setText(translate("mytreeview", "Export as CSV"))



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging

from bitcoin_usb.address_types import SimplePubKeyProvider

from bitcoin_safe.gui.qt.custom_edits import (
    AnalyzerMessage,
    AnalyzerState,
    BaseAnalyzer,
)

logger = logging.getLogger(__name__)

from typing import Callable

import bdkpython as bdk
from bitcoin_qr_tools.data import convert_slip132_to_bip32, is_slip132
from bitcoin_qr_tools.multipath_descriptor import (
    MultipathDescriptor as BitcoinQRMultipathDescriptor,
)
from PyQt6.QtCore import QObject

from ...keystore import KeyStore
from .util import Message


class KeyOriginAnalyzer(BaseAnalyzer, QObject):
    def __init__(self, get_expected_key_origin: Callable[[], str], parent: QObject | None) -> None:
        BaseAnalyzer.__init__(self)
        QObject.__init__(self, parent=parent)
        self.get_expected_key_origin = get_expected_key_origin

    def analyze(self, input: str, pos: int = 0) -> AnalyzerMessage:
        if not input:
            return AnalyzerMessage(self.tr("Missing Key origin"), AnalyzerState.Invalid)

        try:
            input = SimplePubKeyProvider.format_key_origin(input)
        except Exception as e:
            return AnalyzerMessage(str(e), AnalyzerState.Invalid)

        if input == self.get_expected_key_origin():
            return AnalyzerMessage("Expected Key Origin", AnalyzerState.Valid)
        else:
            return AnalyzerMessage(self.tr("Unexpected key origin"), AnalyzerState.Warning)


class FingerprintAnalyzer(BaseAnalyzer, QObject):
    def analyze(self, input: str, pos: int = 0) -> AnalyzerMessage:
        if not input:
            return AnalyzerMessage(self.tr("Missing Fingerprint"), AnalyzerState.Invalid)

        try:
            input = SimplePubKeyProvider.format_fingerprint(input)
        except Exception as e:
            return AnalyzerMessage(str(e), AnalyzerState.Invalid)

        if KeyStore.is_fingerprint_valid(input):
            return AnalyzerMessage("Valid Fingerprint", AnalyzerState.Valid)
        else:
            return AnalyzerMessage(self.tr("Invalid Fingerprint"), AnalyzerState.Invalid)


class XpubAnalyzer(BaseAnalyzer, QObject):
    def __init__(self, network: bdk.Network, parent: QObject | None) -> None:
        BaseAnalyzer.__init__(self)
        QObject.__init__(self, parent=parent)

        self.network = network

    def analyze(self, input: str, pos: int = 0) -> AnalyzerMessage:
        if not input:
            return AnalyzerMessage(self.tr("Missing xPub"), AnalyzerState.Invalid)

        if is_slip132(input):
            Message(
                self.tr("The xpub is in SLIP132 format. Converting to standard format."),
                title=self.tr("Converting format"),
            )
            try:
                input = convert_slip132_to_bip32(input)
            except:
                pass

        if KeyStore.is_xpub_valid(input, network=self.network):
            return AnalyzerMessage("Valid xpub", AnalyzerState.Valid)
        else:
            return AnalyzerMessage(self.tr("Invalid xpub"), AnalyzerState.Invalid)


class SeedAnalyzer(BaseAnalyzer, QObject):
    def analyze(self, input: str, pos: int = 0) -> AnalyzerMessage:
        if not input:
            return AnalyzerMessage(self.tr("Missing Seed"), AnalyzerState.Valid)

        if KeyStore.is_seed_valid(input):
            return AnalyzerMessage("Valid seed", AnalyzerState.Valid)
        else:
            return AnalyzerMessage(self.tr("Invalid seed"), AnalyzerState.Invalid)


class DescriptorAnalyzer(BaseAnalyzer, QObject):
    def __init__(self, network: bdk.Network, parent: QObject | None) -> None:
        BaseAnalyzer.__init__(self)
        QObject.__init__(self, parent=parent)

        self.network = network

    def analyze(self, input: str, pos: int = 0) -> AnalyzerMessage:
        if not input:
            return AnalyzerMessage(self.tr("Missing Descriptor"), AnalyzerState.Invalid)

        if BitcoinQRMultipathDescriptor.is_valid(input, network=self.network):
            return AnalyzerMessage("Valid Descriptor", AnalyzerState.Valid)
        else:
            return AnalyzerMessage(self.tr("Invalid Descriptor"), AnalyzerState.Invalid)


class AddressAnalyzer(BaseAnalyzer, QObject):
    def __init__(self, network: bdk.Network, parent: QObject | None) -> None:
        BaseAnalyzer.__init__(self)
        QObject.__init__(self, parent=parent)

        self.network = network

    def analyze(self, input: str, pos: int = 0) -> AnalyzerMessage:
        if not input:
            return AnalyzerMessage(self.tr("Missing Address"), AnalyzerState.Invalid)

        is_valid = False
        try:
            bdk_address = bdk.Address(input, network=self.network)
            is_valid = bool(bdk_address)
        except:
            is_valid = False

        if is_valid:
            return AnalyzerMessage(self.tr("Valid Address"), AnalyzerState.Valid)
        else:
            return AnalyzerMessage(self.tr("Invalid Address"), AnalyzerState.Invalid)
